"""
  A collection of wrappers for optimisng a function.
  -- kandasamy@cs.cmu.edu
"""

# pylint: disable=invalid-name
# pylint: disable=no-member

from __future__ import print_function
from __future__ import division

from argparse import Namespace
from datetime import datetime
from warnings import warn
import os
import numpy as np
# Optimal Transport
try:
  import ot as py_opt_transport # Python Optimal Transport
except ImportError:
  py_opt_transport = None
# Local imports
try:
  from .direct_fortran import direct as direct_ft_wrap
except ImportError as e:
  fortran_err_msg = ('Could not import Fortran direct library. Dragonfly can still be ' +
                     'used, but might be slightly slower. To get rid of this warning, ' +
                     'install a numpy compatible Fortran compiler (e.g. gfortran) and ' +
                     'the python-dev package and reinstall Dragonfly.')
  warn('%s\n%s'%(e, fortran_err_msg))
  direct_ft_wrap = None
from .general_utils import map_to_bounds
from .doo import DOOFunction, pdoo_wrap


# Optimal transport and Earth mover's distance ===========================================
def opt_transport(supply, demand, costs):
  """ A wrapper for the EMD computation using the Optimal Transport (ot) package.
      if emd_only is False, it only returns the emd value. Else it returns the transport
      matrix and the minimum value of the objective.
  """
  if py_opt_transport is None:
    raise Exception('Could not import Python optimal transport library. ' +
                    'You can install this via "pip install cython POT".')
  supply = supply.astype(np.float64)
  demand = demand.astype(np.float64)
  tot_supply = supply.sum()
  tot_demand = demand.sum()
  supply = supply / tot_supply
  demand = demand / tot_demand
  # Now solve the problem
  T = py_opt_transport.emd(supply, demand, costs)
  T = tot_supply * T
  min_val = np.sum(T * costs)
  emd = min_val/tot_supply
  return T, min_val, emd

# Random sampling -------------------------------
def random_sample(obj, bounds, max_evals, vectorised=True):
  """ Optimises a function by randomly sampling and choosing its maximum. """
  dim = len(bounds)
  rand_pts = map_to_bounds(np.random.random((int(max_evals), dim)), bounds)
  if vectorised:
    obj_vals = obj(rand_pts)
  else:
    obj_vals = np.array([obj(x) for x in rand_pts])
  return rand_pts, obj_vals

# Random maximisation
def random_maximise(obj, bounds, max_evals, return_history=False, vectorised=True):
  """ Optimises a function by randomly sampling and choosing its maximum. """
  rand_pts, obj_vals = random_sample(obj, bounds, max_evals, vectorised)
  max_idx = obj_vals.argmax()
  max_val = obj_vals[max_idx]
  max_pt = rand_pts[max_idx]
  if return_history:
    history = Namespace(query_vals=obj_vals, query_points=rand_pts)
  else:
    history = None
  return max_val, max_pt, history

# Random minimisation
def random_minimise(obj, bounds, max_evals, return_history=False, vectorised=True):
  """ Optimises a function by randomly sampling and choosing its maximum. """
  rand_pts, obj_vals = random_sample(obj, bounds, max_evals, vectorised)
  min_idx = obj_vals.argmin()
  min_val = obj_vals[min_idx]
  min_pt = rand_pts[min_idx]
  if return_history:
    history = Namespace(query_vals=obj_vals, query_points=rand_pts)
  else:
    history = None
  return min_val, min_pt, history

# DIRECT -----------------------------------------
# Some constants
_MAX_DIRECT_FN_EVALS = 2.6e6 # otherwise the fortran software complains

def direct_ft_minimise(obj, bounds, max_evals,
                       return_history=False,
                       eps=1e-5,
                       max_iterations=None,
                       alg_method=0,
                       fglobal=-1e100,
                       fglper=0.01,
                       volper=-1.0,
                       sigmaper=-1.0,
                       log_file_name='',
                       results_file_name='',
                       vectorised=False,
                       alternative_if_direct_not_loaded='pdoo',
                      ):
  """
    A wrapper for the fortran implementation. The four mandatory arguments are self
    explanatory. If return_history is True it also returns the history of evaluations.
    max_iterations is the maximum number of iterations of the direct algorithm.
    I am not sure what the remaining arguments are for.
  """
  # pylint: disable=too-many-locals
  # pylint: disable=too-many-arguments
  if direct_ft_wrap is None:
    report_str = 'Attempted to use direct, but fortran library could not be imported. '
    if alternative_if_direct_not_loaded is None:
      report_str += 'Alternative not specified. Raising exception.'
      raise Exception(report_str)
    elif alternative_if_direct_not_loaded.lower().startswith('rand'):
      report_str += 'Using random optimiser instead of direct.'
      warn(report_str)
      return random_minimise(obj, bounds, max_evals, vectorised)
    elif alternative_if_direct_not_loaded.lower().startswith('pdoo'):
      report_str += 'Using PDOO optimiser instead of direct.'
      warn(report_str)
      return pdoo_minimise(obj, bounds, max_evals)
    else:
      report_str += 'Unknown option for alternative_if_direct_not_loaded: %s'%( \
                     alternative_if_direct_not_loaded)
      raise ValueError(report_str)

  # Preliminaries.
  max_evals = min(_MAX_DIRECT_FN_EVALS, max_evals) # otherwise the fortran sw complains.
  max_iterations = max_evals if max_iterations is None else max_iterations
  bounds = np.array(bounds, dtype=np.float64)
  lower_bounds = bounds[:, 0]
  upper_bounds = bounds[:, 1]
  if len(lower_bounds) != len(upper_bounds):
    raise ValueError('The dimensionality of the lower and upper bounds should match.')

  # Create a wrapper to comply with the fortran requirements.
  def _objective_wrap(x, *_):
    """ A wrapper to comply with the fortran requirements. """
    return (obj(x), 0)

  # Some dummy data to comply with the fortran requirements.
  iidata = np.ones(0, dtype=np.int32)
  ddata = np.ones(0, dtype=np.float64)
  cdata = np.ones([0, 40], dtype=np.uint8)

  # Set up results_file_name if necessary
  if return_history:
    results_file_name = (results_file_name if results_file_name else
                         'direct_results_%s'%(datetime.now().strftime('%m%d-%H%M%S')))
  # Call the function.
  min_pt, min_val, _ = direct_ft_wrap.direct(_objective_wrap,
                                             eps,
                                             max_evals,
                                             max_iterations,
                                             lower_bounds,
                                             upper_bounds,
                                             alg_method,
                                             log_file_name,
                                             results_file_name,
                                             fglobal,
                                             fglper,
                                             volper,
                                             sigmaper,
                                             iidata,
                                             ddata,
                                             cdata
                                            )
  if return_history:
    history = get_direct_history(results_file_name)
    os.remove(results_file_name)
  else:
    history = None
  # return
  return min_val, min_pt, history


def direct_ft_maximise(obj, bounds, max_evals, *args, **kwargs):
  """
    A wrapper for maximising a function which calls direct_ft_minimise. See arguments
    under direct_ft_minimise for more details.
  """
  min_obj = lambda x: -obj(x)
  min_val, max_pt, history = direct_ft_minimise(min_obj, bounds, max_evals,
                                                *args, **kwargs)
  max_val = - min_val
  if history is not None:
    history.curr_opt_vals = -history.curr_opt_vals
    history.curr_true_opt_vals = history.curr_opt_vals
    history.query_vals = -history.query_vals
  return max_val, max_pt, history


def get_direct_history(results_file_name):
  """ Reads the optimisation history from the direct results file. """
  curr_opt_vals = []
  query_vals = []
  query_points = []
  results_file_handle = open(results_file_name, 'r')
  for line in results_file_handle.readlines():
    words = line.strip().split()
    if words[0].isdigit():
      curr_opt_vals.append(float(words[1]))
      query_vals.append(float(words[2]))
      query_points.append([float(w) for w in words[3:]])
  results_file_handle.close()
  return Namespace(curr_opt_vals=np.array(curr_opt_vals),
                   query_vals=np.array(query_vals),
                   query_points=np.array(query_points),
                   curr_true_opt_vals=np.array(curr_opt_vals))


def get_history_from_direct_log(log_file_name):
  """ Returns the history from the direct log file. """
  saved_iterations = [0]
  saved_max_vals = [-np.inf]
  phase = 'boiler'
  log_file_handle = open(log_file_name, 'r')
  for line in log_file_handle.readlines():
    words = line.strip().split()
    if phase == 'boiler':
      if words[0] == 'Iteration':
        phase = 'results'
    elif phase == 'results':
      if len(words) == 3 and words[0].isdigit():
        saved_iterations.append(int(words[1]))
        saved_max_vals.append(-float(words[2]))
      else:
        phase = 'final'
    elif phase == 'final':
      if words[0] == 'Final':
        saved_max_vals.append(max(-float(words[-1]), saved_max_vals[-1]))
        # doing max as the fortran library rounds off the last result for some reason.
      if words[0] == 'Number':
        saved_iterations.append(int(words[-1]))
  log_file_handle.close()
  # Now fill in the rest of the history.
  curr_opt_vals = np.zeros((saved_iterations[-1]), dtype=np.float64)
  for i in range(len(saved_iterations)-1):
    curr_opt_vals[saved_iterations[i]:saved_iterations[i+1]] = saved_max_vals[i]
  curr_opt_vals[-1] = saved_max_vals[-1]
  return Namespace(curr_opt_vals=curr_opt_vals)
# DIRECT end --------------------------------------------


def pdoo_maximise(obj, bounds, max_evals):
  """
    A wrapper for maximising a function which calls doo_wrap. See arguments
    under utils.doo_wrap for more details.
  """
  K = 2
  tol = 1e-3
  nu_max = 1.0
  C_init = 0.8
  rho_max = 0.9
  POO_mult = 0.5
  doo_obj = DOOFunction(obj, bounds)
  max_val, max_pt, history = pdoo_wrap(doo_obj, max_evals, nu_max, rho_max, K,
                                       C_init, tol, POO_mult)
  return max_val, max_pt, history


def pdoo_minimise(obj, bounds, max_evals):
  """
    A wrapper for maximising a function which calls doo_wrap. See arguments
    under utils.doo_wrap for more details.
  """
  max_obj = lambda x: -obj(x)
  max_val, min_pt, history = pdoo_maximise(max_obj, bounds, max_evals)
  min_val = -max_val
  return min_val, min_pt, history


# Utilities for sampling from combined domains ===================================
def latin_hc_indices(dim, num_samples):
  """ Obtains indices for Latin Hyper-cube sampling. """
  index_set = [list(range(num_samples))] * dim
  lhs_indices = []
  for i in range(num_samples):
    curr_idx_idx = np.random.randint(num_samples-i, size=dim)
    curr_idx = [index_set[j][curr_idx_idx[j]] for j in range(dim)]
    index_set = [index_set[j][:curr_idx_idx[j]] + index_set[j][curr_idx_idx[j]+1:]
                 for j in range(dim)]
    lhs_indices.append(curr_idx)
  return lhs_indices

def latin_hc_sampling(dim, num_samples):
  """ Latin Hyper-cube sampling in the unit hyper-cube. """
  if num_samples == 0:
    return np.zeros((0, dim))
  elif num_samples == 1:
    return 0.5 * np.ones((1, dim))
  lhs_lower_boundaries = (np.linspace(0, 1, num_samples+1)[:num_samples]).reshape(1, -1)
  width = lhs_lower_boundaries[0][1] - lhs_lower_boundaries[0][0]
  lhs_lower_boundaries = np.repeat(lhs_lower_boundaries, dim, axis=0).T
  lhs_indices = latin_hc_indices(dim, num_samples)
  lhs_sample_boundaries = []
  for i in range(num_samples):
    curr_idx = lhs_indices[i]
    curr_sample_boundaries = [lhs_lower_boundaries[curr_idx[j]][j] for j in range(dim)]
    lhs_sample_boundaries.append(curr_sample_boundaries)
  lhs_sample_boundaries = np.array(lhs_sample_boundaries)
  uni_random_width = width * np.random.random((num_samples, dim))
  lhs_samples = lhs_sample_boundaries + uni_random_width
  return lhs_samples

def random_sample_from_euclidean_domain(bounds, num_samples, sample_type='rand'):
  """ Samples from a Euclidean Domain. """
  if sample_type == 'rand':
    ret = map_to_bounds(np.random.random((int(num_samples), len(bounds))), bounds)
  elif sample_type == 'latin_hc':
    ret = map_to_bounds(latin_hc_sampling(len(bounds), num_samples), bounds)
  else:
    raise ValueError('Unknown sample_type %s.'%(sample_type))
  return list(ret)

def random_sample_from_discrete_euclidean_domain(valid_vectors, num_samples,
                                                 sample_type='rand'):
  """ Samples from a Euclidean Domain. """
  if sample_type == 'rand' or True:
    ret = valid_vectors[np.random.randint(len(valid_vectors), size=(num_samples, )), :]
  else:
    raise ValueError('Unknown sample_type %s.'%(sample_type))
  return list(ret)

def random_sample_from_integral_domain(bounds, num_samples, sample_type='rand'):
  """ Samples from a Integral Domain. """
  ret = random_sample_from_euclidean_domain(bounds, num_samples, sample_type)
  return [x.astype(np.int) for x in ret]

def random_sample_from_discrete_domain(dscr_vals, num_points=None):
  """ Samples from a discrete domain. """
  def _sample_single_point():
    """ Samples a single point. """
    return [np.random.choice(categ) for categ in dscr_vals]
  # Now draw num_points of them
  num_points_to_sample = 1 if num_points is None else num_points
  if len(dscr_vals) == 0:
    ret = [[]] * num_points_to_sample
  else:
    ret = [_sample_single_point() for _ in range(num_points_to_sample)]
  if num_points is None:
    return ret[0]
  else:
    return ret

def random_sample_from_prod_discrete_domain(list_of_list_of_vals, num_samples):
  """ Sample from a product of discrete domains. """
  return random_sample_from_discrete_domain(list_of_list_of_vals, num_samples)

def random_sample_cts_dscr(obj, cts_bounds, dscr_vals, max_evals, vectorised=True):
  """ Sample from a joint continuous and discrete space. """
  dim = len(cts_bounds)
  cts_rand_pts = map_to_bounds(np.random.random((int(max_evals), dim)), cts_bounds)
  dscr_rand_pts = random_sample_from_discrete_domain(dscr_vals, max_evals)
  if vectorised:
    obj_vals = obj(cts_rand_pts, dscr_rand_pts)
  else:
    obj_vals = np.array([obj(cx, dx) for (cx, dx) in zip(cts_rand_pts, dscr_rand_pts)])
  return cts_rand_pts, dscr_rand_pts, obj_vals

# Random maximisation
def random_maximise_cts_dscr(obj, cts_bounds, dscr_vals, max_evals,
                             return_history=False, vectorised=True):
  """ Optimises a function by randomly sampling and choosing its maximum. """
  cts_rand_pts, dscr_rand_pts, obj_vals = random_sample_cts_dscr(obj, cts_bounds, \
                                                         dscr_vals, max_evals, vectorised)
  max_idx = obj_vals.argmax()
  max_val = obj_vals[max_idx]
  max_cts_pt = cts_rand_pts[max_idx]
  max_dscr_pt = dscr_rand_pts[max_idx]
  if return_history:
    history = Namespace(query_vals=obj_vals, query_cts_points=cts_rand_pts,
                        query_dscr_points=dscr_rand_pts)
  else:
    history = None
  return max_val, max_cts_pt, max_dscr_pt, history




"""
  Harness for GP Bandit Optimisation.
  -- kandasamy@cs.cmu.edu
"""
from __future__ import division

# pylint: disable=invalid-name
# pylint: disable=no-member
# pylint: disable=redefined-builtin
# pylint: disable=unbalanced-tuple-unpacking

from argparse import Namespace
import numpy as np
# Local imports
from ..exd import domains
from ..exd.cp_domain_utils import get_processed_func_from_raw_func_for_cp_domain, \
                                load_cp_domain_from_config_file, load_config_file
from ..exd.exd_core import mf_exd_args
from ..exd.exd_utils import get_euclidean_initial_qinfos, get_cp_domain_initial_qinfos
from ..exd.experiment_caller import CPFunctionCaller, get_multifunction_caller_from_config
from ..gp.euclidean_gp import EuclideanGPFitter, euclidean_gp_args, \
                            EuclideanMFGPFitter, euclidean_mf_gp_args
from ..gp.cartesian_product_gp import cartesian_product_gp_args, \
                                    cartesian_product_mf_gp_args, \
                                    CPGPFitter, CPMFGPFitter
from . import gpb_acquisitions
from .blackbox_optimiser import blackbox_opt_args, BlackboxOptimiser, \
                                   CalledMFOptimiserWithSFCaller
from ..utils.general_utils import block_augment_array, get_idxs_from_list_of_lists
from ..utils.option_handler import get_option_specs, load_options
from ..utils.reporters import get_reporter


gp_bandit_args = [ \
  get_option_specs('acq', False, 'default', \
    'Which acquisition to use: ts, ucb, ei, ttei, bucb. If using multiple ' + \
    'give them as a hyphen separated list e.g. ucb-ts-ei-ttei'),
  get_option_specs('acq_probs', False, 'adaptive', \
    'With what probability should we choose each strategy given in acq. If "uniform" ' + \
    'we we will use uniform probabilities and if "adaptive" we will use adaptive ' + \
    'probabilities which weight acquisitions according to how well they do.'),
  get_option_specs('acq_opt_method', False, 'default', \
    'Which optimiser to use when maximising the acquisition function.'),
  get_option_specs('handle_parallel', False, 'halluc', \
    'How to handle parallelisations. Should be halluc or naive.'),
  get_option_specs('acq_opt_max_evals', False, -1, \
    'Number of evaluations when maximising acquisition. If negative uses default value.'),
  # The following are for managing GP hyper-parameters. They override hp_tune_criterion
  # and ml_hp_tune_opt from the GP args.
  get_option_specs('gpb_hp_tune_criterion', False, 'ml-post_sampling',
                   'Which criterion to use when tuning hyper-parameters. Other ' +
                   'options are post_sampling and post_mean.'),
  get_option_specs('gpb_hp_tune_probs', False, '0.3-0.7', \
    'With what probability should we choose each strategy given in hp_tune_criterion.' + \
    'If "uniform" we we will use uniform probabilities and if "adaptive" we will use ' + \
    'adaptive probabilities which weight acquisitions according to how well they do.'),
  get_option_specs('gpb_ml_hp_tune_opt', False, 'default',
                   'Which optimiser to use when maximising the tuning criterion.'),
  get_option_specs('gpb_post_hp_tune_method', False, 'slice',
                   'Which sampling to use when maximising the tuning criterion. Other ' +
                   'option is nuts'),
  get_option_specs('gpb_post_hp_tune_burn', False, -1,
                   'How many initial samples to ignore during sampling.'),
  get_option_specs('gpb_post_hp_tune_offset', False, 25,
                   'How many samples to ignore between samples.'),
  get_option_specs('rand_exp_sampling_replace', False, False, \
    'Whether to replace already sampled values or not in rand_exp_sampling.'),
  # For multi-fidelity BO
  get_option_specs('mf_strategy', False, 'boca',
                   'Which multi-fidelity strategy to use. Should be one of {boca}.'),
  # The following are perhaps not so important. Some have not been implemented yet.
  get_option_specs('shrink_kernel_with_time', False, 0,
                   'If True, shrinks the kernel with time so that we don\'t get stuck.'),
  get_option_specs('perturb_thresh', False, 1e-4, \
    ('If the next point chosen is too close to an exisiting point by this times the ' \
     'diameter, then we will perturb the point a little bit before querying. This is ' \
     'mainly to avoid numerical stability issues.')),
  get_option_specs('track_every_time_step', False, 0,
                   ('If 1, it tracks every time step.')),
  get_option_specs('next_pt_std_thresh', False, 0.005,
    ('If the std of the queried point queries below this times the kernel scale ', \
     'frequently we will reduce the bandwidth range')),
  # Miscellanneous
  get_option_specs('nn_report_results_every', False, 1,
    ('If NN variables are present, report results more frequently')),
  ]

mf_gp_bandit_args = [ \
  get_option_specs('target_fidel_to_opt_query_frac_max', False, 0.5,
                   ('A target to maintain on the number of queries to fidel_to_opt.')),
  get_option_specs('target_fidel_to_opt_query_frac_min', False, 0.25,
                   ('A target to maintain on the number of queries to fidel_to_opt.')),
  get_option_specs('boca_thresh_window_length', False, 20, \
    ('The window length to keep checking if the target fidel_to_opt is achieved.')),
  get_option_specs('boca_thresh_coeff_init', False, 0.01,
                   ('The coefficient to use in determining the threshold for boca.')),
  get_option_specs('boca_thresh_multiplier', False, 1.1, \
    ('The amount by which to multiply/divide the threshold coeff for boca.')),
  get_option_specs('boca_max_low_fidel_cost_ratio', False, 0.90, \
    ('If the fidel_cost_ratio is larger than this, just query at fidel_to_opt.')), \
  ]

euclidean_specific_gp_bandit_args = [ \
  get_option_specs('euc_init_method', False, 'latin_hc', \
    'Method to obtain initial queries. Is used if get_initial_qinfos is None.'), \
  ]

def get_all_gp_bandit_args(additional_args):
  """ Returns the GP bandit arguments from the arguments for the GP. """
  return additional_args + blackbox_opt_args + gp_bandit_args

def get_all_mf_gp_bandit_args(additional_args):
  """ Returns the GP bandit arguments from the arguments for the GP. """
  return additional_args + blackbox_opt_args + gp_bandit_args + mf_exd_args + \
         mf_gp_bandit_args

def get_all_euc_gp_bandit_args(additional_args=None):
  """ Returns all GP bandit arguments. """
  if additional_args is None:
    additional_args = []
  return get_all_gp_bandit_args(additional_args) + euclidean_gp_args + \
         euclidean_specific_gp_bandit_args

def get_all_mf_euc_gp_bandit_args(additional_args=None):
  """ Returns all GP bandit arguments. """
  if additional_args is None:
    additional_args = []
  return get_all_mf_gp_bandit_args(additional_args) + euclidean_mf_gp_args + \
         euclidean_specific_gp_bandit_args

def get_all_cp_gp_bandit_args(additional_args=None):
  """ Returns all Cartesian Product GP bandit arguments. """
  if additional_args is None:
    additional_args = []
  return get_all_gp_bandit_args(additional_args) + cartesian_product_gp_args

def get_all_mf_cp_gp_bandit_args(additional_args=None):
  """ Returns all Cartesian Product GP bandit arguments. """
  if additional_args is None:
    additional_args = []
  return get_all_mf_gp_bandit_args(additional_args) + cartesian_product_mf_gp_args

def get_default_acquisition_for_domain(domain):
  """ Returns the default acquisition for the domain. """
  if domain.get_type() == 'euclidean':
    return 'ei-ucb-ttei-ts-add_ucb'
  else:
    return 'ei-ucb-ttei-ts'

def get_default_acq_opt_method_for_domain(domain):
  """ Returns the default acquisition optimisation method for the domain. """
  if domain.get_type() == 'euclidean':
    if domain.get_dim() > 60:
      return 'pdoo'
    else:
      return 'direct'
  elif domain.get_type() == 'cartesian_product':
    if all([dom.get_type() == 'euclidean' for dom in domain.list_of_domains]) and \
       (not domain.has_constraints()):
      if domain.get_dim() > 60:
        return 'pdoo'
      else:
        return 'direct'
    else:
      return 'ga'


# The GPBandit Class
# ========================================================================================
class GPBandit(BlackboxOptimiser):
  """ GPBandit Class. """
  # pylint: disable=attribute-defined-outside-init

  # Constructor.
  def __init__(self, func_caller, worker_manager, is_mf=False,
               options=None, reporter=None):
    """ Constructor. """
    self._is_mf = is_mf
    if is_mf and not func_caller.is_mf():
      raise CalledMFOptimiserWithSFCaller(self, func_caller)
    super(GPBandit, self).__init__(func_caller, worker_manager, None,
                                   options=options, reporter=reporter)

  def is_an_mf_method(self):
    """ Returns Truee since this is a MF method. """
    return self._is_mf

  def _get_method_str(self):
    """ Returns a string describing the method. """
    gpb_str = 'mfbo-%s'%(self.options.mf_strategy) if self.is_an_mf_method() else 'bo'
    return '%s(%s)'%(gpb_str, '-'.join(self.acqs_to_use))

  def _opt_method_set_up(self):
    """ Some set up for the GPBandit class. """
    # Set up acquisition optimisation
    self.gp = None
    # Set up for acquisition optimisation and then acquisition
    self._set_up_acq_opt()
    self._set_up_for_acquisition()
    # Override options for hp_tune_criterion and ml_hp_tune_opt
    self.options.hp_tune_criterion = self.options.gpb_hp_tune_criterion
    self.options.hp_tune_probs = self.options.gpb_hp_tune_probs
    self.options.ml_hp_tune_opt = self.options.gpb_ml_hp_tune_opt
    self.options.post_hp_tune_method = self.options.gpb_post_hp_tune_method
    self.options.post_hp_tune_burn = self.options.gpb_post_hp_tune_burn
    self.options.post_hp_tune_offset = self.options.gpb_post_hp_tune_offset
    # To store in history
    self.history.query_acqs = []
    self.to_copy_from_qinfo_to_history['curr_acq'] = 'query_acqs'
    # For multi-fidelity
    if self.is_an_mf_method():
      self.mf_params_for_anc_data = {}
      if self.options.mf_strategy == 'boca':
        self.mf_params_for_anc_data['boca_thresh_coeff'] = \
          self.options.boca_thresh_coeff_init
        self.mf_params_for_anc_data['boca_max_low_fidel_cost_ratio'] = \
          self.options.boca_max_low_fidel_cost_ratio
    self._child_opt_method_set_up()

  def _set_up_for_acquisition(self):
    """ Set up for acquisition. """
    if self.options.acq == 'default':
      acq = self._get_default_acquisition_for_domain(self.domain)
    else:
      acq = self.options.acq
    self.acqs_to_use = [elem.lower() for elem in acq.split('-')]
    self.acqs_to_use_counter = {key: 0 for key in self.acqs_to_use}
    if self.options.acq_probs == 'uniform':
      self.acq_probs = np.ones(len(self.acqs_to_use)) / float(len(self.acqs_to_use))
    elif self.options.acq_probs == 'adaptive':
      self.acq_uniform_sampling_prob = 0.05
      self.acq_sampling_weights = {key: 1.0 for key in self.acqs_to_use}
      self.acq_probs = self._get_adaptive_ensemble_acq_probs()
    else:
      self.acq_probs = np.array([float(x) for x in self.options.acq_probs.split('-')])
    self.acq_probs = self.acq_probs / self.acq_probs.sum()
    assert len(self.acq_probs) == len(self.acqs_to_use)

  @classmethod
  def _get_default_acquisition_for_domain(cls, domain):
    """ Return default acqusition for domain. """
    return get_default_acquisition_for_domain(domain)

  def _child_opt_method_set_up(self):
    """ Set up for child class. Override this method in child class"""
    pass

  def _get_adaptive_ensemble_acq_probs(self):
    """ Computes the adaptive ensemble acqusitions probs. """
    num_acqs = len(self.acqs_to_use)
    uniform_sampling_probs = self.acq_uniform_sampling_prob * \
                             np.ones((num_acqs,)) / num_acqs
    acq_succ_counter = np.array([self.acq_sampling_weights[key] for
                                 key in self.acqs_to_use])
    acq_use_counter = np.array([self.acqs_to_use_counter[key] for
                                key in self.acqs_to_use])
    acq_weights = acq_succ_counter / np.sqrt(1 + acq_use_counter)
    acq_norm_weights = acq_weights / acq_weights.sum()
    adaptive_sampling_probs = (1 - self.acq_uniform_sampling_prob) * acq_norm_weights
    ret = uniform_sampling_probs + adaptive_sampling_probs
    return ret / ret.sum()

  def _set_up_acq_opt(self):
    """ Sets up optimisation for acquisition. """
    # First set up function to get maximum evaluations.
    if isinstance(self.options.acq_opt_max_evals, int):
      if self.options.acq_opt_max_evals > 0:
        self.get_acq_opt_max_evals = lambda t: self.options.acq_opt_max_evals
      else:
        self.get_acq_opt_max_evals = None
    else: # In this case, the user likely passed a function here.
      self.get_acq_opt_max_evals = self.options.acq_opt_max_evals
    # Additional set up based on the specific optimisation procedure
    if self.options.acq_opt_method == 'default':
      acq_opt_method = get_default_acq_opt_method_for_domain(self.domain)
    else:
      acq_opt_method = self.options.acq_opt_method
    self.acq_opt_method = acq_opt_method
    self._domain_specific_acq_opt_set_up()

  def _opt_method_update_history(self, qinfo):
    """ Update history for GP bandit specific statistics. """
    if hasattr(qinfo, 'curr_acq'):
      self.acqs_to_use_counter[qinfo.curr_acq] += 1
      if self.options.acq_probs == 'adaptive' and \
         (len(self.history.curr_opt_vals) >= 2 and
          self.history.curr_opt_vals[-1] > self.history.curr_opt_vals[-2]):
        self.acq_sampling_weights[qinfo.curr_acq] += 1
    if hasattr(self, 'gp_processor') and hasattr(qinfo, 'hp_tune_method') and \
       (len(self.history.curr_opt_vals) >= 2 and \
        self.history.curr_opt_vals[-1] > self.history.curr_opt_vals[-2]):
      self.gp_processor.gp_fitter.update_hp_tune_method_weight(qinfo.hp_tune_method)
    self._child_opt_method_update_history(qinfo)

  def _child_opt_method_update_history(self, qinfo):
    """ Update history for child GP bandit specific statistics. """
    pass

  def _domain_specific_acq_opt_set_up(self):
    """ Set up acquisition optimisation for the child class. """
    raise NotImplementedError('Implement in a child class.')

  # Managing the GP ---------------------------------------------------------
  def _set_next_gp(self):
    """ Returns the next GP. """
    if not hasattr(self, 'gp_processor') or self.gp_processor is None:
      self._build_new_gp()
    ret = self.gp_processor.gp_fitter.get_next_gp()
    self.gp_processor.fit_type = ret[0]
    self.gp_processor.hp_tune_method = ret[1]
    self.gp = ret[2]
    self._domain_specific_set_next_gp()
    if self.gp_processor.fit_type in ['sample_hps_with_probs', \
                                      'post_sample_hps_with_probs']:
      reg_data = self._get_gp_reg_data()
      self._child_set_gp_data(reg_data)
#     if self.step_idx == self.last_model_build_at or \
#        self.step_idx == self.last_model_build_at + self.worker_manager.num_workers:
#       # printing multiple times every model build is a way to check that post_sampling
#       # is picking different hyperparameters.
#       self._report_current_gp()
    # We need to do this separately since posterior sampling complicates how we keep
    # track of the GPs.
    if self.step_idx == self.last_model_build_at and \
       self.options.report_model_on_each_build:
      self._report_current_gp()

  def _domain_specific_set_next_gp(self):
    """ Sets the next GP in child class """
    pass

  def _child_set_gp_data(self, reg_data):
    """ Set data in child. """
    if self.is_an_mf_method():
      self.gp.set_mf_data(reg_data[0], reg_data[1], reg_data[2], build_posterior=True)
    else:
      self.gp.set_data(reg_data[0], reg_data[1], build_posterior=True)

  def _child_build_new_model(self):
    """ Builds a new model. """
    self._build_new_gp()

  def _report_model(self):
    """ Report the current model. """
    # We will do this separately since posterior sampling complicates how we keep
    # track of the GPs. See _set_next_gp()
    pass

  def _report_current_gp(self):
    """ Reports the current GP. """
    gp_fit_report_str = '    -- GP at iter %d: %s'%(self.step_idx, str(self.gp))
    self.reporter.writeln(gp_fit_report_str)

  def _get_opt_method_header_str(self):
    """ Header for optimisation Method. """
    return ', acqs=<num_times_each_acquisition_was_used>'

  def _get_opt_method_report_results_str(self):
    """ Any details to include in a child method when reporting results.
    """
    ret_list = ['%s:%d'%(key, self.acqs_to_use_counter[key]) for key in self.acqs_to_use]
    ret = 'acqs=[' + ', '.join(ret_list) + ']'
    return ', ' + ret

  def _get_gp_reg_data(self):
    """ Returns the current data collected using initialisation and optimisation.
        If not multi-fidelity returns a 2-tuple (X,Y) where X are in the domain and Y
        are the outputs. If multi-fidelity, returns a 3-tuple (Z,X,Y) where Z is a list
        of fidelities.
    """
    reg_X_raw = self.prev_eval_points + self.history.query_points
    reg_Y_raw = self.prev_eval_vals + self.history.query_vals
    finite_idxs = [idx for idx in range(len(reg_Y_raw)) if np.isfinite(reg_Y_raw[idx])]
    reg_X = [reg_X_raw[idx] for idx in finite_idxs]
    reg_Y = [reg_Y_raw[idx] for idx in finite_idxs]
    if self.is_an_mf_method():
      reg_Z_raw = self.prev_eval_fidels + self.history.query_fidels
      reg_Z = [reg_Z_raw[idx] for idx in finite_idxs]
      return reg_Z, reg_X, reg_Y
    else:
      return reg_X, reg_Y

  def _get_gp_fitter(self, reg_data, use_additive=False):
    """ Returns a GP Fitter. """
    if self.is_an_mf_method():
      return self._get_mf_gp_fitter(reg_data, use_additive)
    else:
      return self._get_non_mf_gp_fitter(reg_data, use_additive)

  def _get_mf_gp_fitter(self, reg_data, use_additive=False):
    """ Returns the Multi-fidelity GP Fitter. Can be overridded by a child class. """
    raise NotImplementedError('Implement in a Child class.')

  def _get_non_mf_gp_fitter(self, reg_data, use_additive=False):
    """ Returns the NOn-Multi-fidelity GP Fitter. Can be overridded by a child class. """
    raise NotImplementedError('Implement in a Child class.')

  def _build_new_gp(self):
    """ Builds a GP with the data in history and stores in self.gp. """
    if hasattr(self.func_caller, 'init_gp') and self.func_caller.init_gp is not None:
      # If you know the true GP.
      raise NotImplementedError('Not implemented passing given GP yet.')
    else:
      if self.options.shrink_kernel_with_time:
        raise NotImplementedError('Not implemented kernel shrinking for the GP yet.')
      # Invoke the GP fitter.
      reg_data = self._get_gp_reg_data()
      gp_fitter = self._get_gp_fitter(reg_data)
      # Fits gp and adds it to gp_processor
      gp_fitter.fit_gp_for_gp_bandit(self.options.build_new_model_every)
      self.gp = None # Mostly to avoid bugs
      self.gp_processor = Namespace()
      self.gp_processor.gp_fitter = gp_fitter
      self._domain_specific_build_new_gp(reg_data)

  def _domain_specific_build_new_gp(self, reg_data):
    """ Builds an extra GP in child class if required. """
    pass

  def _add_data_to_model(self, qinfos):
    """ Add data to self.gp """
    if len(qinfos) == 0:
      return
    new_points = [qinfo.point for qinfo in qinfos]
    new_vals = [qinfo.val for qinfo in qinfos]
    if self.is_an_mf_method():
      new_fidels = [qinfo.fidel for qinfo in qinfos]
      self._add_data_to_gp((new_fidels, new_points, new_vals))
    else:
      self._add_data_to_gp((new_points, new_vals))

  def _add_data_to_gp(self, new_data):
    """ Adds data to the GP. """
    # Add data to the GP only if we will be repeating with the same GP.
    if hasattr(self, 'gp_processor') and hasattr(self.gp_processor, 'fit_type') and \
      self.gp_processor.fit_type == 'fitted_gp':
      if self.is_an_mf_method():
        self.gp.add_mf_data_multiple(new_data[0], new_data[1], new_data[2])
      else:
        self.gp.add_data_multiple(new_data[0], new_data[1])
    self._child_add_data_to_gp(new_data)

  def _child_add_data_to_gp(self, new_data):
    """ Adds data from the child class to the GP. """
    pass

  # Methods needed for optimisation ------------------------------------------
  def _get_next_acq(self):
    """ Gets the acquisition to use in the current iteration. """
    if self.options.acq_probs == 'adaptive':
      self.acq_probs = self._get_adaptive_ensemble_acq_probs()
    ret = np.random.choice(self.acqs_to_use, p=self.acq_probs)
    return ret

  def _get_ancillary_data_for_acquisition(self, curr_acq):
    """ Returns ancillary data for the acquisitions. """
    max_num_acq_opt_evals = self.get_acq_opt_max_evals(self.step_idx)
    ret = Namespace(curr_acq=curr_acq,
                    max_evals=max_num_acq_opt_evals,
                    t=self.step_idx,
                    domain=self.domain,
                    curr_max_val=self.curr_opt_val,
                    eval_points_in_progress=self.eval_points_in_progress,
                    acq_opt_method=self.acq_opt_method,
                    handle_parallel=self.options.handle_parallel,
                    mf_strategy=self.options.mf_strategy,
                    is_mf=self.is_an_mf_method(),
                   )
    if curr_acq == 'add_ucb':
      ret.domain_bounds = self.domain.bounds
    if self.is_an_mf_method():
      for key, value in list(self.mf_params_for_anc_data.items()):
        setattr(ret, key, value)
      ret.eval_fidels_in_progress = self.eval_fidels_in_progress
      ret.eval_fidel_points_in_progress = self.gp.get_ZX_from_ZZ_XX( \
                              self.eval_fidels_in_progress, self.eval_points_in_progress)
    return ret

  def _determine_next_query(self):
    """ Determine the next point for evaluation. """
    curr_acq = self._get_next_acq()
    anc_data = self._get_ancillary_data_for_acquisition(curr_acq)
    select_pt_func = getattr(gpb_acquisitions.asy, curr_acq)
    qinfo = Namespace(curr_acq=curr_acq,
                      hp_tune_method=self.gp_processor.hp_tune_method)
    if self.is_an_mf_method():
      if self.options.mf_strategy == 'boca':
        next_eval_fidel, next_eval_point = gpb_acquisitions.boca(select_pt_func, \
          self.gp, anc_data, self.func_caller)
        qinfo.fidel = next_eval_fidel
        qinfo.point = next_eval_point
      else:
        raise ValueError('Unknown mf_strategy: %s.'%(self.options.mf_strategy))
    else:
      next_eval_point = select_pt_func(self.gp, anc_data)
      qinfo.point = next_eval_point
    return qinfo

  def _determine_next_batch_of_queries(self, batch_size):
    """ Determine the next batch of evaluation points. """
    curr_acq = self._get_next_acq()
    anc_data = self._get_ancillary_data_for_acquisition(curr_acq)
    select_pt_func = getattr(gpb_acquisitions.syn, curr_acq)
    if self.is_an_mf_method():
      raise NotImplementedError('Not Implemented synchronous mf yet!')
    else:
      next_batch_of_eval_points = select_pt_func(batch_size, self.gp, anc_data)
      qinfos = [Namespace(point=pt,
                          hp_tune_method=self.gp_processor.hp_tune_method,
                          curr_acq=curr_acq) for pt in next_batch_of_eval_points]
    return qinfos

  def _main_loop_pre_boca(self):
    """ Things to be done before each iteration of the optimisation loop for BOCA. """
    last_window_queries_at_f2o = self.history.query_at_fidel_to_opts[ \
                                                 -self.options.boca_thresh_window_length:]
    last_window_f2o_frac = sum(last_window_queries_at_f2o) / \
                              float(self.options.boca_thresh_window_length)
    if last_window_f2o_frac <= self.options.target_fidel_to_opt_query_frac_min:
      self.mf_params_for_anc_data['boca_thresh_coeff'] *= \
        self.options.boca_thresh_multiplier
    elif last_window_f2o_frac >= self.options.target_fidel_to_opt_query_frac_max:
      self.mf_params_for_anc_data['boca_thresh_coeff'] /= \
        self.options.boca_thresh_multiplier
    if len(self.history.query_vals) > 1:
      self.mf_params_for_anc_data['y_range'] = \
        max(self.history.query_vals) - min(self.history.query_vals)
    else:
      self.mf_params_for_anc_data['y_range'] = 1.0

  def _main_loop_pre(self):
    """ Things to be done before each iteration of the optimisation loop. """
    self._set_next_gp() # set the next GP
    # For BOCA
    if self.is_an_mf_method():
      if self.options.mf_strategy == 'boca':
        self._main_loop_pre_boca()
# GP Bandit class ends here ==========================================================


# A Euclidean GP Bandit
class EuclideanGPBandit(GPBandit):
  """ A GP Bandit for Euclidean Spaces. """

  # Constructor.
  def __init__(self, func_caller, worker_manager, is_mf=False,
               options=None, reporter=None):
    """ Constructor. """
    if options is None:
      reporter = get_reporter(reporter)
      if is_mf:
        all_args = get_all_mf_euc_gp_bandit_args()
      else:
        all_args = get_all_euc_gp_bandit_args()
      options = load_options(all_args, reporter=reporter)
    super(EuclideanGPBandit, self).__init__(func_caller, worker_manager, is_mf=is_mf,
                                            options=options, reporter=reporter)

  def _get_mf_gp_fitter(self, reg_data, use_additive=False):
    """ Returns the Multi-fidelity GP Fitter. Can be overridded by a child class. """
    options = Namespace(**vars(self.options))
    if use_additive:
      options.domain_use_additive_gp = use_additive
    if use_additive and options.domain_kernel_type == 'esp':
      options.domain_kernel_type = options.domain_esp_kernel_type
    return EuclideanMFGPFitter(reg_data[0], reg_data[1], reg_data[2],
                               options=options, reporter=self.reporter)

  def _get_non_mf_gp_fitter(self, reg_data, use_additive=False):
    """ Returns the NOn-Multi-fidelity GP Fitter. Can be overridded by a child class. """
    options = Namespace(**vars(self.options))
    if use_additive:
      options.use_additive_gp = use_additive
    if use_additive and options.kernel_type == 'esp':
      options.kernel_type = options.esp_kernel_type
    return EuclideanGPFitter(reg_data[0], reg_data[1],
                             options=options, reporter=self.reporter)

  def _child_opt_method_set_up(self):
    """ Some set up for the EuclideanGPBandit class. """
    # Override init method
    self.options.init_method = self.options.euc_init_method
    self.add_gp = None
    # Flag for creating extra add_gp
    self.req_add_gp = False
    if self.is_an_mf_method():
      if not self.options.domain_use_additive_gp and 'add_ucb' in self.acqs_to_use:
        self.req_add_gp = True
    else:
      if not self.options.use_additive_gp and 'add_ucb' in self.acqs_to_use:
        self.req_add_gp = True

  def _domain_specific_build_new_gp(self, reg_data):
    """ Builds an additive GP if required and stores it in self.add_gp. """
    if self.req_add_gp:
      add_gp_fitter = self._get_gp_fitter(reg_data, use_additive=True)
      # Fits gp and adds it to add_gp_processor
      add_gp_fitter.fit_gp_for_gp_bandit(self.options.build_new_model_every)
      self.add_gp = None # Mostly to avoid bugs
      self.add_gp_processor = Namespace()
      self.add_gp_processor.gp_fitter = add_gp_fitter

  def _child_add_data_to_gp(self, new_data):
    """ Adds data from the child class to the EuclideanGP """
    if hasattr(self, 'add_gp_processor') and  hasattr(self.add_gp_processor, 'fit_type') \
      and self.add_gp_processor.fit_type == 'fitted_gp':
      if self.is_an_mf_method():
        if self.add_gp is not None:
          self.add_gp.add_mf_data_multiple(new_data[0], new_data[1], new_data[2])
      else:
        if self.add_gp is not None:
          self.add_gp.add_data_multiple(new_data[0], new_data[1])

  def _child_opt_method_update_history(self, qinfo):
    """ Update history for EuclideanGP bandit specific statistics. """
    if hasattr(self, 'add_gp_processor') and hasattr(qinfo, 'hp_tune_method') and \
       (len(self.history.curr_opt_vals) >= 2 and \
       self.history.curr_opt_vals[-1] > self.history.curr_opt_vals[-2]):
      self.add_gp_processor.gp_fitter.update_hp_tune_method_weight(qinfo.hp_tune_method)

  def _domain_specific_set_next_gp(self):
    if hasattr(self, 'add_gp_processor'):
      ret = self.add_gp_processor.gp_fitter.get_next_gp()
      self.add_gp_processor.fit_type = ret[0]
      self.add_gp_processor.hp_tune_method = ret[1]
      self.add_gp = ret[2]
      if self.add_gp_processor.fit_type in ['sample_hps_with_probs',
                                            'post_sample_hps_with_probs']:
        reg_data = self._get_gp_reg_data()
        self._set_add_gp_data(reg_data)

  def _set_add_gp_data(self, reg_data):
    """ Set data in child. Can be overridden by a child class. """
    if self.is_an_mf_method():
      self.add_gp.set_mf_data(reg_data[0], reg_data[1], reg_data[2], build_posterior=True)
    else:
      self.add_gp.set_data(reg_data[0], reg_data[1], build_posterior=True)

  def _determine_next_query(self):
    """ Determine the next point for evaluation. """
    curr_acq = self._get_next_acq()
    anc_data = self._get_ancillary_data_for_acquisition(curr_acq)
    select_pt_func = getattr(gpb_acquisitions.asy, curr_acq)
    if curr_acq == 'add_ucb':
      qinfo_hp_tune_method = self.add_gp_processor.hp_tune_method
    else:
      qinfo_hp_tune_method = self.gp_processor.hp_tune_method
    qinfo = Namespace(curr_acq=curr_acq, hp_tune_method=qinfo_hp_tune_method)
    if self.is_an_mf_method():
      if self.options.mf_strategy == 'boca':
        if self.add_gp is None or curr_acq != 'add_ucb':
          next_eval_fidel, next_eval_point = gpb_acquisitions.boca(select_pt_func, \
            self.gp, anc_data, self.func_caller)
        else:
          next_eval_fidel, next_eval_point = gpb_acquisitions.boca(select_pt_func, \
            self.add_gp, anc_data, self.func_caller)
        qinfo.fidel = next_eval_fidel
        qinfo.point = next_eval_point
      else:
        raise ValueError('Unknown mf_strategy: %s.'%(self.options.mf_strategy))
    else:
      if self.add_gp is None or curr_acq != 'add_ucb':
        next_eval_point = select_pt_func(self.gp, anc_data)
      else:
        next_eval_point = select_pt_func(self.add_gp, anc_data)
      qinfo.point = next_eval_point
    return qinfo

  def _determine_next_batch_of_queries(self, batch_size):
    """ Determine the next batch of evaluation points. """
    curr_acq = self._get_next_acq()
    anc_data = self._get_ancillary_data_for_acquisition(curr_acq)
    select_pt_func = getattr(gpb_acquisitions.syn, curr_acq)
    if curr_acq == 'add_ucb':
      qinfo_hp_tune_method = self.add_gp_processor.hp_tune_method
    else:
      qinfo_hp_tune_method = self.gp_processor.hp_tune_method
    if self.is_an_mf_method():
      if self.options.mf_strategy == 'boca':
        # TODO: implement this.
        raise NotImplementedError('Not Implemented Synchronous Boca yet!')
      else:
        raise ValueError('Unknown mf_strategy: %s.'%(self.options.mf_strategy))
    else:
      if self.add_gp is None or curr_acq != 'add_ucb':
        next_batch_of_eval_points = select_pt_func(batch_size, self.gp, anc_data)
      else:
        next_batch_of_eval_points = select_pt_func(batch_size, self.add_gp, anc_data)
      qinfos = [Namespace(point=pt, hp_tune_method=qinfo_hp_tune_method)
                for pt in next_batch_of_eval_points]
    return qinfos

  def _get_initial_qinfos(self, num_init_evals):
    """ Returns initial qinfos. """
    if self.is_an_mf_method():
      return get_euclidean_initial_qinfos(self.options.init_method, num_init_evals, \
               self.domain.bounds, self.options.fidel_init_method, \
               self.fidel_space.bounds, self.func_caller.fidel_to_opt, \
               self.options.init_set_to_fidel_to_opt_with_prob)
    else:
      return get_euclidean_initial_qinfos(self.options.init_method, num_init_evals,
                                          self.domain.bounds)

  def _domain_specific_acq_opt_set_up(self):
    """ Set up acquisition optimisation for the child class. """
    if self.acq_opt_method.lower() in ['direct']:
      self._set_up_euc_acq_opt_direct()
    elif self.acq_opt_method.lower() in ['pdoo']:
      self._set_up_euc_acq_opt_pdoo()
    elif self.acq_opt_method.lower() == 'rand':
      self._set_up_euc_acq_opt_rand()
    else:
      raise NotImplementedError('Not implemented acquisition optimisation for %s yet.'%( \
                                self.acq_opt_method))

  # Any of these set up methods can be overridden by a child class -------------------
  def _set_up_euc_acq_opt_direct(self):
    """ Sets up optimisation for acquisition using direct/pdoo. """
    if self.get_acq_opt_max_evals is None:
      lead_const = 1 * min(5, self.domain.get_dim())**2
      self.get_acq_opt_max_evals = lambda t: np.clip(lead_const * np.sqrt(min(t, 1000)),
                                                     1000, 3e4)

  def _set_up_euc_acq_opt_pdoo(self):
    """ Sets up optimisation for acquisition using direct/pdoo. """
    if self.get_acq_opt_max_evals is None:
      lead_const = 2 * min(5, self.domain.get_dim())**2
      self.get_acq_opt_max_evals = lambda t: np.clip(lead_const * np.sqrt(min(t, 1000)),
                                                     2000, 6e4)

  def _set_up_euc_acq_opt_rand(self):
    """ Sets up optimisation for acquisition using random search. """
    if self.get_acq_opt_max_evals is None:
      lead_const = 10 * min(5, self.domain.get_dim())**2
      self.get_acq_opt_max_evals = lambda t: np.clip(lead_const * np.sqrt(min(t, 1000)),
                                                     2000, 3e4)


class CPGPBandit(GPBandit):
  """ A GP Bandit class on Cartesian product spaces. """

  def __init__(self, func_caller, worker_manager, is_mf=False,
               domain_dist_computers=None, options=None, reporter=None):
    """ Constructor. """
    if options is None:
      reporter = get_reporter(reporter)
      if is_mf:
        all_args = get_all_mf_euc_gp_bandit_args()
      else:
        all_args = get_all_cp_gp_bandit_args()
      options = load_options(all_args, reporter)
    self.domain_dist_computers = domain_dist_computers
    super(CPGPBandit, self).__init__(func_caller, worker_manager, is_mf=is_mf,
                                     options=options, reporter=reporter)

  def _child_opt_method_set_up(self):
    """ Set up for child class. Override this method in child class. """
    self.domain_lists_of_dists = None
    if self.domain_dist_computers is None:
      self.domain_dist_computers = [None] * self.domain.num_domains
    self.kernel_params_for_each_domain = [{} for _ in range(self.domain.num_domains)]
    # Create a Dummy GP Fitter so that we can get the mislabel and struct coeffs for
    # otmann.
    if self.is_an_mf_method():
      fs_orderings = self.func_caller.fidel_space_orderings
      d_orderings = self.func_caller.domain_orderings
      dummy_gp_fitter = CPMFGPFitter([], [], [], config=None,
        fidel_space=self.func_caller.fidel_space,
        domain=self.func_caller.domain,
        fidel_space_kernel_ordering=fs_orderings.kernel_ordering,
        domain_kernel_ordering=d_orderings.kernel_ordering,
        fidel_space_lists_of_dists=None,
        domain_lists_of_dists=None,
        fidel_space_dist_computers=None,
        domain_dist_computers=None,
        options=self.options, reporter=self.reporter)
    else:
      dummy_gp_fitter = CPGPFitter([], [], self.func_caller.domain,
         domain_kernel_ordering=self.func_caller.domain_orderings.kernel_ordering,
         domain_lists_of_dists=None,
         domain_dist_computers=None,
         options=self.options, reporter=self.reporter)
    # Pre-compute distances for all sub-domains in domain - not doing for fidel_space
    # since we don't expect pre-computing distances will be necessary there.
    for idx, dom in enumerate(self.domain.list_of_domains):
      if dom.get_type() == 'neural_network' and self.domain_dist_computers[idx] is None:
        from ..nn.otmann import get_otmann_distance_computer_from_args
        otm_mislabel_coeffs =  \
          dummy_gp_fitter.domain_kernel_params_for_each_domain[idx].otmann_mislabel_coeffs
        otm_struct_coeffs =  \
          dummy_gp_fitter.domain_kernel_params_for_each_domain[idx].otmann_struct_coeffs
        self.domain_dist_computers[idx] = get_otmann_distance_computer_from_args(
          dom.nn_type, self.options.otmann_non_assignment_penalty,
          otm_mislabel_coeffs, otm_struct_coeffs, self.options.otmann_dist_type)
        self.kernel_params_for_each_domain[idx]['otmann_dist_type'] = \
          self.options.otmann_dist_type
    # Report more frquently if Neural networks are present
    domain_types = [dom.get_type() for dom in self.domain.list_of_domains]
    if 'neural_network' in domain_types:
      self.options.report_results_every = self.options.nn_report_results_every

  def _domain_specific_acq_opt_set_up(self):
    """ Set up acquisition optimisation for the child class. """
    if self.acq_opt_method.lower() in ['direct']:
      self._set_up_cp_acq_opt_direct()
    elif self.acq_opt_method.lower() in ['pdoo']:
      self._set_up_cp_acq_opt_pdoo()
    elif self.acq_opt_method.lower() == 'rand':
      self._set_up_cp_acq_opt_rand()
    elif self.acq_opt_method.lower().startswith('ga'):
      self._set_up_cp_acq_opt_ga()
    else:
      raise ValueError('Unrecognised acq_opt_method "%s".'%(self.acq_opt_method))

  # Any of these set up methods can be overridden by a child class -------------------
  def _set_up_cp_acq_opt_with_params(self, lead_const, min_iters, max_iters):
    """ Set up acquisition optimisation with params. """
    if self.get_acq_opt_max_evals is None:
      dim_factor = lead_const * min(5, self.domain.get_dim())**2
      self.get_acq_opt_max_evals = lambda t: np.clip(dim_factor * np.sqrt(min(t, 1000)),
                                                     min_iters, max_iters)

  def _set_up_cp_acq_opt_direct(self):
    """ Sets up optimisation for acquisition using direct/pdoo. """
    self._set_up_cp_acq_opt_with_params(2, 1000, 3e4)

  def _set_up_cp_acq_opt_pdoo(self):
    """ Sets up optimisation for acquisition using direct/pdoo. """
    self._set_up_cp_acq_opt_with_params(2, 2000, 6e4)

  def _set_up_cp_acq_opt_rand(self):
    """ Set up optimisation for acquisition using rand. """
    self._set_up_cp_acq_opt_with_params(1, 1000, 3e4)

  def _set_up_cp_acq_opt_ga(self):
    """ Set up optimisation for acquisition using rand. """
    domain_types = [dom.get_type() for dom in self.domain.list_of_domains]
    if 'neural_network' in domain_types:
      # Because Neural networks can be quite expensive
      self._set_up_cp_acq_opt_with_params(1, 300, 1e3)
    else:
      self._set_up_cp_acq_opt_with_params(1, 1000, 3e4)

  def _compute_lists_of_dists(self, X1, X2):
    """ Computes lists of dists. """
    ret = [None] * self.domain.num_domains
    for idx, dist_comp in enumerate(self.domain_dist_computers):
      if dist_comp is not None:
        X1_idx = get_idxs_from_list_of_lists(X1, idx)
        X2_idx = X1_idx if X1 is X2 else get_idxs_from_list_of_lists(X2, idx)
        ret[idx] = dist_comp(X1_idx, X2_idx)
    return ret

  def _add_data_to_gp(self, new_data):
    """ Adds data to the GP. Overriding this method. """
    # First add it to the list of dists
    if self.is_an_mf_method():
      _, new_reg_X, _ = new_data
    else:
      new_reg_X, _ = new_data
    if self.domain_lists_of_dists is None:
      # First time, so use all the data
      self.domain_lists_of_dists = self._compute_lists_of_dists(new_reg_X, new_reg_X)
      self.already_evaluated_dists_for = new_reg_X
    else:
      domain_lists_of_dists_new_new = self._compute_lists_of_dists(new_reg_X, new_reg_X)
      domain_lists_of_dists_old_new = self._compute_lists_of_dists(
                                 self.already_evaluated_dists_for, new_reg_X)
      for i in range(self.domain.num_domains): # through each domain
        if self.domain_lists_of_dists[i] is None:
          continue
        for j in range(len(domain_lists_of_dists_new_new[i])):
          # iterate through each dist in curr dom
          self.domain_lists_of_dists[i][j] = block_augment_array(
            self.domain_lists_of_dists[i][j], domain_lists_of_dists_old_new[i][j],
            domain_lists_of_dists_old_new[i][j].T, domain_lists_of_dists_new_new[i][j])
      self.already_evaluated_dists_for.extend(new_reg_X)
    # Add data to the GP as we will be repeating with the same GP.
    if hasattr(self, 'gp_processor') and hasattr(self.gp_processor, 'fit_type') and \
      self.gp_processor.fit_type == 'fitted_gp':
      reg_data = self._get_gp_reg_data()
      if self.is_an_mf_method():
        self.gp.set_mf_data(reg_data[0], reg_data[1], reg_data[2], build_posterior=False)
        self.gp.set_domain_lists_of_dists(self.domain_lists_of_dists)
      else:
        self.gp.set_data(reg_data[0], reg_data[1], build_posterior=False)
        self.gp.set_domain_lists_of_dists(self.domain_lists_of_dists)
      # Build the posterior
      self.gp.build_posterior()

  def _get_initial_qinfos(self, num_init_evals):
    """ Returns initial qinfos. """
    if self.is_an_mf_method():
      return get_cp_domain_initial_qinfos(self.domain, num_init_evals,
        fidel_space=self.fidel_space, fidel_to_opt=self.func_caller.fidel_to_opt,
        set_to_fidel_to_opt_with_prob=self.options.init_set_to_fidel_to_opt_with_prob,
        dom_euclidean_sample_type='latin_hc',
        dom_integral_sample_type='latin_hc',
        dom_nn_sample_type='rand',
        fidel_space_euclidean_sample_type='latin_hc',
        fidel_space_integral_sample_type='latin_hc',
        fidel_space_nn_sample_type='rand')
    else:
      return get_cp_domain_initial_qinfos(self.domain, num_init_evals,
                                          dom_euclidean_sample_type='latin_hc',
                                          dom_integral_sample_type='latin_hc',
                                          dom_nn_sample_type='rand')

  def _get_mf_gp_fitter(self, reg_data, use_additive=False):
    """ Returns the Multi-fidelity GP Fitter. Can be overridded by a child class. """
    # We are not maintaining a list of distances for the domain or the fidelity space.
    fs_orderings = self.func_caller.fidel_space_orderings
    return CPMFGPFitter(reg_data[0], reg_data[1], reg_data[2], config=None,
             fidel_space=self.func_caller.fidel_space,
             domain=self.func_caller.domain,
             fidel_space_kernel_ordering=fs_orderings.kernel_ordering,
             domain_kernel_ordering=self.func_caller.domain_orderings.kernel_ordering,
             fidel_space_lists_of_dists=None,
             domain_lists_of_dists=self.domain_lists_of_dists,
             fidel_space_dist_computers=None,
             domain_dist_computers=self.domain_dist_computers,
             options=self.options, reporter=self.reporter)

  def _get_non_mf_gp_fitter(self, reg_data, use_additive=False):
    """ Returns the NOn-Multi-fidelity GP Fitter. Can be overridded by a child class. """
    return CPGPFitter(reg_data[0], reg_data[1], self.func_caller.domain,
             domain_kernel_ordering=self.func_caller.domain_orderings.kernel_ordering,
             domain_lists_of_dists=self.domain_lists_of_dists,
             domain_dist_computers=self.domain_dist_computers,
             options=self.options, reporter=self.reporter)


# APIs for Euclidean GP Bandit optimisation. ---------------------------------------------
# 1. Optimisation from a FunctionCaller object.
def gpb_from_func_caller(func_caller, worker_manager, max_capital, is_mf, mode=None,
                         acq=None, mf_strategy=None, domain_add_max_group_size=-1,
                         options=None, reporter='default'):
  """ GP Bandit optimisation from a utils.function_caller.FunctionCaller instance.
    domain_add_max_group_size indicates whether we should use an additive model or not.
    If its negative, then use a non-additive model. If its positive, then use an additive
    model with maximum group size as given. If zero, then use the default in options.
  """
  # pylint: disable=too-many-branches
  reporter = get_reporter(reporter)
  # Decide which optimiser to use.
  if is_mf:
    if isinstance(func_caller.fidel_space, domains.EuclideanDomain) and \
       isinstance(func_caller.domain, domains.EuclideanDomain):
      optimiser_constructor = EuclideanGPBandit
      dflt_list_of_options = get_all_mf_euc_gp_bandit_args()
    elif isinstance(func_caller.fidel_space, domains.CartesianProductDomain) and \
         isinstance(func_caller.domain, domains.CartesianProductDomain):
      optimiser_constructor = CPGPBandit
      dflt_list_of_options = get_all_mf_cp_gp_bandit_args()
    else:
      raise ValueError(
        'GP Bandit not implemented for fidel_space, domain of type %s, %s.'%(
        func_caller.fidel_space.get_type(), func_caller.domain.get_type()))
  else:
    if isinstance(func_caller.domain, domains.EuclideanDomain):
      optimiser_constructor = EuclideanGPBandit
      dflt_list_of_options = get_all_euc_gp_bandit_args()
    elif isinstance(func_caller.domain, domains.CartesianProductDomain):
      optimiser_constructor = CPGPBandit
      dflt_list_of_options = get_all_cp_gp_bandit_args()
    else:
      raise ValueError('GP Bandit not implemented for domain of type %s.'%( \
                       type(func_caller.domain)))
  # Load options
  if options is None:
    reporter = get_reporter(reporter)
    options = load_options(dflt_list_of_options, reporter=reporter)
  if acq is not None:
    options.acq = acq
  if mode is not None:
    options.mode = mode
  if mf_strategy is not None:
    options.mf_strategy = mf_strategy
  from ..exd.worker_manager import RealWorkerManager, SyntheticWorkerManager
  if isinstance(worker_manager, RealWorkerManager):
    options.capital_type = 'realtime'
  elif isinstance(worker_manager, SyntheticWorkerManager):
    options.capital_type = 'return_value'
  # Domain Specific Parameters
  if isinstance(func_caller.domain, domains.EuclideanDomain) \
    and domain_add_max_group_size >= 0:
    # TODO (KK): this domain_add_max_group_size parameter is legacy and has no effect.
    if is_mf:
      options.domain_use_additive_gp = True
      if domain_add_max_group_size > 0:
        options.domain_add_max_group_size = domain_add_max_group_size
    else:
      options.use_additive_gp = True
      if domain_add_max_group_size > 0:
        options.add_max_group_size = domain_add_max_group_size
  # create optimiser and return
  optimiser = optimiser_constructor(func_caller, worker_manager, is_mf=is_mf,
                                    options=options, reporter=reporter)
  return optimiser.optimise(max_capital)


def cp_gpb_from_raw_args(raw_func, domain_config_file, *args, **kwargs):
  """ GP Bandit from the raw_func and domain configuration file. """
  cp_dom, orderings = load_cp_domain_from_config_file(domain_config_file)
  proc_func = get_processed_func_from_raw_func_for_cp_domain(
                raw_func, cp_dom, orderings.index_ordering, orderings.dim_ordering)
  func_caller = CPFunctionCaller(proc_func, cp_dom, raw_func=raw_func,
                                 domain_orderings=orderings)
  return gpb_from_func_caller(func_caller, *args, **kwargs)


def mf_cp_gpb_from_raw_args(raw_func, raw_fidel_cost_func, domain_config_file,
                            *args, **kwargs):
  """ Multi-fidelity GP Bandit from raw_functions and domain configuration file. """
  config = load_config_file(domain_config_file)
  func_caller = get_multifunction_caller_from_config(raw_func, config,
                               raw_fidel_cost_func=raw_fidel_cost_func)
  return gpb_from_func_caller(func_caller, *args, **kwargs)


# Alternative names
bo_from_func_caller = gpb_from_func_caller




"""
  A manager for multiple workers.
  -- kandasamy@cs.cmu.edu
"""
from __future__ import print_function
from __future__ import division

# pylint: disable=invalid-name
# pylint: disable=abstract-class-not-used
# pylint: disable=abstract-class-little-used

from argparse import Namespace
from multiprocessing import Process
import numpy as np
import os
import pickle
import shutil
import time
try:
  from sets import Set
except ImportError:
  Set = set

# Local
from .exd_utils import EVAL_ERROR_CODE

_TIME_TOL = 1e-5


class WorkerManager(object):
  """ A Base class for a worker manager. """

  def __init__(self, worker_ids):
    """ Constructor. """
    if hasattr(worker_ids, '__iter__'):
      self.worker_ids = worker_ids
    else:
      self.worker_ids = list(range(worker_ids))
    self.num_workers = len(self.worker_ids)
    # These will be set in reset
    self.experiment_designer = None
    self.latest_results = None
    # Reset
    self.reset()

  def reset(self):
    """ Resets everything. """
    self.experiment_designer = None
    self.latest_results = [] # A list of namespaces
    self._child_reset()

  def _child_reset(self):
    """ Child reset. """
    raise NotImplementedError('Implement in a child class.')

  def fetch_latest_results(self):
    """ Returns the latest results. """
    ret_idxs = []
    for i in range(len(self.latest_results)):
      if (self.latest_results[i].receive_time <=
            self.experiment_designer.get_curr_spent_capital() + _TIME_TOL):
        ret_idxs.append(i)
    keep_idxs = [i for i in range(len(self.latest_results)) if i not in ret_idxs]
    ret = [self.latest_results[i] for i in ret_idxs]
    self.latest_results = [self.latest_results[i] for i in keep_idxs]
    return ret

  def close_all_queries(self):
    """ Closes all queries. """
    raise NotImplementedError('Implement in a child class.')

  def set_experiment_designer(self, experiment_designer):
    """ Set the experiment designer. """
    self.experiment_designer = experiment_designer

  def a_worker_is_free(self):
    """ Returns true if a worker is free. """
    raise NotImplementedError('Implement in a child class.')

  def all_workers_are_free(self):
    """ Returns true if all workers are free. """
    raise NotImplementedError('Implement in a child class.')

  def _dispatch_experiment(self, func_caller, qinfo, **kwargs):
    """ Dispatches job. """
    raise NotImplementedError('Implement in a child class.')

  def dispatch_single_experiment(self, func_caller, qinfo, **kwargs):
    """ Dispatches job. """
    raise NotImplementedError('Implement in a child class.')

  def dispatch_batch_of_experiments(self, func_caller, qinfos, **kwargs):
    """ Dispatches an entire batch of experiments. """
    raise NotImplementedError('Implement in a child class.')

  def get_time_distro_info(self):
    """ Returns information on the time distribution. """
    #pylint: disable=no-self-use
    return ''

  def get_poll_time_real(self):
    """ Returns the poll time. """
    raise NotImplementedError('Implement in a child class.')


# A synthetic worker manager - for simulating multiple workers ---------------------------
class SyntheticWorkerManager(WorkerManager):
  """ A Worker manager for synthetic functions. Mostly to be used in simulations. """

  def __init__(self, num_workers, time_distro='caller_eval_cost',
               time_distro_params=None):
    """ Constructor. """
    self.worker_pipe = None
    super(SyntheticWorkerManager, self).__init__(num_workers)
    # Set up the time sampler
    self.time_distro = time_distro
    self.time_distro_params = time_distro_params
    self.time_sampler = None
    self._set_up_time_sampler()

  def _set_up_time_sampler(self):
    """ Set up the sampler for the time random variable. """
    self.time_distro_params = Namespace() if self.time_distro_params is None else \
                              self.time_distro_params
    if self.time_distro == 'caller_eval_cost':
      pass
    elif self.time_distro == 'const':
      if not hasattr(self.time_distro_params, 'const_val'):
        self.time_distro_params.const_val = 1
      self.time_sampler = lambda num_samples: (np.ones((num_samples,)) *
                                               self.time_distro_params.const_val)
    elif self.time_distro == 'uniform':
      if not hasattr(self.time_distro_params, 'ub'):
        self.time_distro_params.ub = 2.0
        self.time_distro_params.lb = 0.0
      ub = self.time_distro_params.ub
      lb = self.time_distro_params.lb
      self.time_sampler = lambda num_samples: (np.random.random((num_samples,)) *
                                               (ub - lb) + lb)
    elif self.time_distro == 'halfnormal':
      if not hasattr(self.time_distro_params, 'ub'):
        self.time_distro_params.sigma = np.sqrt(np.pi/2)
      self.time_sampler = lambda num_samples: np.abs(np.random.normal(
        scale=self.time_distro_params.sigma, size=(num_samples,)))
    else:
      raise NotImplementedError('Not implemented time_distro = %s yet.'%(
                                self.time_distro))

  def _child_reset(self):
    """ Child reset. """
    self.worker_pipe = [[wid, 0.0] for wid in self.worker_ids]

  def sort_worker_pipe(self):
    """ Sorts worker pipe by finish time. """
    self.worker_pipe.sort(key=lambda x: x[-1])

  def a_worker_is_free(self):
    """ Returns true if a worker is free. """
    return self.worker_pipe[0][-1] # Always return true as this is synthetic.

  def all_workers_are_free(self):
    """ Returns true if all workers are free. """
    return self.worker_pipe[-1][-1]

  def close_all_queries(self):
    """ Close all queries. """
    pass

  def _dispatch_experiment(self, func_caller, qinfo, worker_id, **kwargs):
    """ Dispatch experiment. """
    # Set worker id and whether or not eval_time should be returned
    qinfo.worker_id = worker_id # indicate which worker
    qinfo = func_caller.eval_from_qinfo(qinfo, **kwargs)
    if self.time_distro == 'caller_eval_cost':
      if hasattr(qinfo, 'caller_eval_cost') and qinfo.caller_eval_cost is not None:
        qinfo.eval_time = qinfo.caller_eval_cost
      else:
        qinfo.eval_time = 1.0
    else:
      qinfo.eval_time = float(self.time_sampler(1))
    qinfo.receive_time = qinfo.send_time + qinfo.eval_time
    # Store the result in latest_results
    self.latest_results.append(qinfo)
    return qinfo

  def dispatch_single_experiment(self, func_caller, qinfo, **kwargs):
    """ Dispatch a single experiment. """
    worker_id = self.worker_pipe[0][0]
    qinfo = self._dispatch_experiment(func_caller, qinfo, worker_id, **kwargs)
    # Sort the pipe
    self.worker_pipe[0][-1] = qinfo.receive_time
    self.sort_worker_pipe()

  def dispatch_batch_of_experiments(self, func_caller, qinfos, **kwargs):
    """ Dispatches an entire batch of experiments. """
    assert len(qinfos) == self.num_workers
    for idx in range(self.num_workers):
      qinfo = self._dispatch_experiment(func_caller, qinfos[idx],
                                             self.worker_pipe[idx][0], **kwargs)
      self.worker_pipe[idx][-1] = qinfo.receive_time
    self.sort_worker_pipe()

  def get_time_distro_info(self):
    """ Returns information on the time distribution. """
    return self.time_distro

  def get_poll_time_real(self):
    """ Return 0.0 as the poll time. """
    return 0.0


# Real worker manager - for simulating multiple workers --------------------------------
class RealWorkerManager(WorkerManager):
  """ A worker manager for resnet. """
  # pylint: disable=attribute-defined-outside-init

  def __init__(self, worker_ids, tmp_dir,
               poll_time=0.5, sleep_time_after_new_process=0.5):
    """ Constructor. """
    super(RealWorkerManager, self).__init__(worker_ids)
    self.poll_time = poll_time
    self.sleep_time_after_new_process = sleep_time_after_new_process
    self.tmp_dir = tmp_dir
    self._rwm_set_up()
    self._child_reset()

  def _rwm_set_up(self):
    """ Sets things up for the child. """
    # Create the result directories. """
    self.result_dir_names = {wid:'%s/result_%s'%(self.tmp_dir, str(wid)) for wid in
                                                 self.worker_ids}
    # Create the working directories
    self.working_dir_names = {wid:'%s/working_%s/tmp'%(self.tmp_dir,
                               str(wid)) for wid in self.worker_ids}
    # Create the last receive times
    self.last_receive_times = {wid:0.0 for wid in self.worker_ids}
    # Create file names
    self._result_file_name = 'result.p'
    self._num_file_read_attempts = 10
#     self._file_read_poll_time = 0.5 # wait for 0.5 seconds

  @classmethod
  def _delete_dirs(cls, list_of_dir_names):
    """ Deletes a list of directories. """
    for dir_name in list_of_dir_names:
      if os.path.exists(dir_name):
        shutil.rmtree(dir_name)

  @classmethod
  def _delete_and_create_dirs(cls, list_of_dir_names):
    """ Deletes a list of directories and creates new ones. """
    for dir_name in list_of_dir_names:
      if os.path.exists(dir_name):
        shutil.rmtree(dir_name)
      os.makedirs(dir_name)

  def _child_reset(self):
    """ Resets child. """
    # Delete/create the result and working directories.
    if not hasattr(self, 'result_dir_names'): # Just for the super constructor.
      return
    self._delete_and_create_dirs(list(self.result_dir_names.values()))
    self._delete_dirs(list(self.working_dir_names.values()))
    self.free_workers = Set(self.worker_ids)
    self.func_callers_for_each_worker = {wid:None for wid in self.worker_ids}
    self.qinfos_in_progress = {wid:None for wid in self.worker_ids}
    self.worker_processes = {wid:None for wid in self.worker_ids}

  def _get_result_file_name_for_worker(self, worker_id):
    """ Computes the result file name for the worker. """
    return os.path.join(self.result_dir_names[worker_id], self._result_file_name)

#   def _read_result_from_file(self, result_file_name):
#     """ Reads the result from the file name. """
#     #pylint: disable=bare-except
#     num_attempts = 0
#     while num_attempts < self._num_file_read_attempts:
#       try:
#         file_reader = open(result_file_name, 'r')
#         read_in = file_reader.read().strip()
#         try:
#           # try converting to float. If not successful, it is likely an error string.
#           read_in = float(read_in)
#         except:
#           pass
#         file_reader.close()
#         result = read_in
#         break
#       except:
#         print('Encountered error when reading %s. Trying again.'%(result_file_name))
#         time.sleep(self.poll_time)
#         file_reader.close()
#         result = EVAL_ERROR_CODE
#     return result

  def _read_result_from_file(self, result_file_name):
    """ Reads the result from the file name. """
    #pylint: disable=bare-except
    num_attempts = 0
    while num_attempts < self._num_file_read_attempts:
      try:
        file_reader = open(result_file_name, 'rb')
        result = pickle.load(file_reader)
        break
      except:
        print('Encountered error when reading %s. Trying again.'%(result_file_name))
        time.sleep(self.poll_time)
        file_reader.close()
        result = EVAL_ERROR_CODE
    return result

  def _read_result_from_worker_and_update(self, worker_id):
    """ Reads the result from the worker. """
    # pylint: disable=maybe-no-member
    # Read the file
    result_file_name = self._get_result_file_name_for_worker(worker_id)
    result_qinfo = self._read_result_from_file(result_file_name)
    saved_qinfo = self.qinfos_in_progress[worker_id]
    # Now update the relevant qinfo and put it to latest_results
    if isinstance(result_qinfo, Namespace):
      assert self.func_callers_for_each_worker[worker_id].domain.members_are_equal(
               result_qinfo.point, saved_qinfo.point)
      qinfo = result_qinfo
    elif result_qinfo == EVAL_ERROR_CODE:
      qinfo = saved_qinfo
      qinfo.val = EVAL_ERROR_CODE
    else:
      raise ValueError('Could not read qinfo object: %s.'%(str(qinfo)))
    qinfo.receive_time = self.experiment_designer.get_curr_spent_capital()
    qinfo.eval_time = qinfo.receive_time - qinfo.send_time
    if not hasattr(qinfo, 'true_val'):
      qinfo.true_val = qinfo.val
    self.latest_results.append(qinfo)
    # Update receive time
    self.last_receive_times[worker_id] = qinfo.receive_time
    # Delete the file.
    os.remove(result_file_name)
    # Delete content in a working directory.
    shutil.rmtree(self.working_dir_names[worker_id])
    # Add the worker to the list of free workers and clear qinfos in progress.
    self.worker_processes[worker_id].terminate()
    self.worker_processes[worker_id] = None
    self.qinfos_in_progress[worker_id] = None
    self.func_callers_for_each_worker[worker_id] = None
    self.free_workers.add(worker_id)

  def _worker_is_free(self, worker_id):
    """ Checks if worker with worker_id is free. """
    if worker_id in self.free_workers:
      return True
    worker_result_file_name = self._get_result_file_name_for_worker(worker_id)
    if os.path.exists(worker_result_file_name):
      self._read_result_from_worker_and_update(worker_id)
    else:
      return False

  def _get_last_receive_time(self):
    """ Returns the last time we received a job. """
    all_receive_times = list(self.last_receive_times.values())
    return max(all_receive_times)

  def a_worker_is_free(self):
    """ Returns true if a worker is free. """
    for wid in self.worker_ids:
      if self._worker_is_free(wid):
        return self._get_last_receive_time()
    return None

  def all_workers_are_free(self):
    """ Returns true if all workers are free. """
    all_are_free = True
    for wid in self.worker_ids:
      all_are_free = self._worker_is_free(wid) and all_are_free
    if all_are_free:
      return self._get_last_receive_time()
    else:
      return None

  def _dispatch_experiment(self, func_caller, qinfo, worker_id, **kwargs):
    """ Dispatches experiment to worker_id. """
    #pylint: disable=star-args
    if self.qinfos_in_progress[worker_id] is not None:
      err_msg = 'qinfos_in_progress: %s,\nfree_workers: %s.'%(
                   str(self.qinfos_in_progress), str(self.free_workers))
      print(err_msg)
      raise ValueError('Check if worker is free before sending experiment.')
    # First add all the data to qinfo
    qinfo.worker_id = worker_id
    qinfo.working_dir = self.working_dir_names[worker_id]
    qinfo.result_file = self._get_result_file_name_for_worker(worker_id)
    # Create the working directory
    os.makedirs(qinfo.working_dir)
    # Dispatch the experiment in a new process
    target_func = lambda: func_caller.eval_from_qinfo(qinfo, **kwargs)
    self.worker_processes[worker_id] = Process(target=target_func)
    self.worker_processes[worker_id].start()
    time.sleep(self.sleep_time_after_new_process)
    # Add the qinfo to the in progress bar and remove from free_workers
    self.qinfos_in_progress[worker_id] = qinfo
    self.func_callers_for_each_worker[worker_id] = func_caller
    self.free_workers.discard(worker_id)

  def dispatch_single_experiment(self, func_caller, qinfo, **kwargs):
    """ Dispatches a single experiment to a free worker. """
    worker_id = self.free_workers.pop()
    self._dispatch_experiment(func_caller, qinfo, worker_id, **kwargs)

  def dispatch_batch_of_experiments(self, func_caller, qinfos, **kwargs):
    """ Dispatches a batch of experiments. """
    assert len(qinfos) == self.num_workers
    for idx in range(self.num_workers):
      self._dispatch_experiment(func_caller, qinfos[idx], self.worker_ids[idx], **kwargs)

  def close_all_queries(self):
    """ Closes all queries. """
    pass

  def get_time_distro_info(self):
    """ Returns information on the time distribution. """
    return 'realtime'

  def get_poll_time_real(self):
    """ Return 0.0 as the poll time. """
    return self.poll_time




"""
  Dragonfly: An open source library for scalable and robust Bayesian optimisation.
  -- kandasamy@cs.cmu.edu
"""
# pylint: disable=relative-import
# pylint: disable=invalid-name
# pylint: disable=maybe-no-member
# pylint: disable=no-member

# Local
from .api_utils import preprocess_multifidelity_arguments, preprocess_arguments, \
                       get_worker_manager_from_capital_type, \
                       post_process_history_for_minimisation, \
                       load_options_for_method
from ..exd.experiment_caller import EuclideanFunctionCaller, CPFunctionCaller
from ..exd.cp_domain_utils import get_raw_from_processed_via_config
from ..opt.cp_ga_optimiser import cp_ga_optimiser_from_proc_args
from ..opt.gp_bandit import gpb_from_func_caller
from ..opt.random_optimiser import random_optimiser_from_func_caller, \
                                   mf_random_optimiser_from_func_caller
from ..utils.oper_utils import direct_ft_maximise
from ..utils.doo import pdoo_maximise_from_args


def maximise_multifidelity_function(func, fidel_space, domain, fidel_to_opt,
  fidel_cost_func, max_capital, capital_type='return_value', opt_method='bo',
  config=None, options=None, reporter='default'):
  # pylint: disable=too-many-arguments
  """
    Maximises a multi-fidelity function 'func' over the domain 'domain' and fidelity
    space 'fidel_space'.
    Inputs:
      func: The function to be maximised. Takes two arguments func(z, x) where z is a
            member of the fidelity space and x is a member of the domain.
      fidel_space: The fidelity space from which the approximations are obtained.
                   Should be an instance of the Domain class in exd/domains.py.
                   If of the form [[l1, u1], [l2, u2], ...] where li < ui, then we will
                   create a Euclidean domain with lower bounds li and upper bounds
                   ui along each dimension.
      domain: The domain over which the function should be maximised, should be an
              instance of the Domain class in exd/domains.py.
              If domain is a list of the form [[l1, u1], [l2, u2], ...] where li < ui,
              then we will create a Euclidean domain with lower bounds li and upper bounds
              ui along each dimension.
      fidel_to_opt: The point at the fidelity space at which we wish to maximise func.
      max_capital: The maximum capital (time budget or number of evaluations) available
                   for optimisation.
      capital_type: The type of capital. Should be one of 'return_value' or 'realtime'.
                    Default is return_value which indicates we will use the value returned
                    by fidel_cost_func. If realtime, we will use wall clock time.
      opt_method: The method used for optimisation. Could be one of bo or rand.
                  Default is bo. bo - Bayesian optimisation, rand - Random search.
      config: Either a configuration file or or parameters returned by
              exd.cp_domain_utils.load_config_file. config can be None only if domain
              is a EuclideanDomain object.
      options: Additional hyper-parameters for optimisation, as a namespace.
      reporter: A stream to print progress made during optimisation, or one of the
                following strings 'default', 'silent'. If 'silent', then it suppresses
                all outputs. If 'default', writes to stdout.
      * Alternatively, domain and fidelity space could be None if config is either a
        path_name to a configuration file or has configuration parameters.
    Returns:
      opt_val: The maximum value found during the optimisation procdure.
      opt_pt: The corresponding optimum point.
      history: A record of the optimisation procedure which include the point evaluated
               and the values at each time step.
  """
  # Preprocess domain and config arguments
  raw_func = func
  fidel_space, domain, preproc_func_list, fidel_cost_func, fidel_to_opt, config, _ = \
    preprocess_multifidelity_arguments(fidel_space, domain, [func], fidel_cost_func,
                                       fidel_to_opt, config)
  func = preproc_func_list[0]
  # Load arguments and function caller
  if fidel_space.get_type() == 'euclidean' and domain.get_type() == 'euclidean':
    func_caller = EuclideanFunctionCaller(func, domain, vectorised=False,
                    raw_fidel_space=fidel_space, fidel_cost_func=fidel_cost_func,
                    raw_fidel_to_opt=fidel_to_opt)
  else:
    func_caller = CPFunctionCaller(func, domain, '', raw_func=raw_func,
      domain_orderings=config.domain_orderings, fidel_space=fidel_space,
      fidel_cost_func=fidel_cost_func, fidel_to_opt=fidel_to_opt,
      fidel_space_orderings=config.fidel_space_orderings)
  # load options
  options = load_options_for_method(opt_method, 'mfopt', domain, capital_type, options)
  # Create worker manager
  worker_manager = get_worker_manager_from_capital_type(capital_type)

  # Select method here -----------------------------------------------------------
  if opt_method == 'bo':
    opt_val, opt_pt, history = gpb_from_func_caller(func_caller, worker_manager,
                                       max_capital, is_mf=True, options=options,
                                       reporter=reporter)
  elif opt_method == 'rand':
    opt_val, opt_pt, history = mf_random_optimiser_from_func_caller(func_caller,
                worker_manager, max_capital, options=options, reporter=reporter)

  # Post processing
  if domain.get_type() == 'euclidean' and config is None:
    opt_pt = func_caller.get_raw_domain_coords(opt_pt)
    history.curr_opt_points = [func_caller.get_raw_domain_coords(pt)
                               for pt in history.curr_opt_points]
    history.query_points = [func_caller.get_raw_domain_coords(pt)
                            for pt in history.query_points]
    history.query_fidels = [func_caller.get_raw_fidel_coords(fidel)
                            for fidel in history.query_fidels]
  else:
    def _get_raw_from_processed_for_mf(fidel, pt):
      """ Returns raw point from processed point by accounting for the fact that a
          point could be None in the multi-fidelity setting. """
      if fidel is None or pt is None:
        return None, None
      else:
        return get_raw_from_processed_via_config((fidel, pt), config)
    # Now re-write curr_opt_points
    opt_pt = _get_raw_from_processed_for_mf(fidel_to_opt, opt_pt)[1]
    history.curr_opt_points_raw = [_get_raw_from_processed_for_mf(fidel_to_opt, pt)[1]
                                   for pt in history.curr_opt_points]
    query_fidel_points_raw = [_get_raw_from_processed_for_mf(fidel, pt)
      for fidel, pt in zip(history.query_fidels, history.query_points)]
    history.query_fidels = [zx[0] for zx in query_fidel_points_raw]
    history.query_points = [zx[1] for zx in query_fidel_points_raw]
  return opt_val, opt_pt, history


def maximise_function(func, domain, max_capital, capital_type='num_evals',
                      opt_method='bo', config=None, options=None, reporter='default'):
  """
    Maximises a function 'func' over the domain 'domain'.
    Inputs:
      func: The function to be maximised.
      domain: The domain over which the function should be maximised, should be an
              instance of the Domain class in exd/domains.py.
              If domain is a list of the form [[l1, u1], [l2, u2], ...] where li < ui,
              then we will create a Euclidean domain with lower bounds li and upper bounds
              ui along each dimension.
      max_capital: The maximum capital (time budget or number of evaluations) available
                   for optimisation.
      capital_type: The type of capital. Should be one of 'return_value' or 'realtime'.
                    Default is return_value which indicates we will use the value returned
                    by fidel_cost_func. If realtime, we will use wall clock time.
      opt_method: The method used for optimisation. Could be one of bo, rand, ga, ea,
                  direct, or pdoo. Default is bo.
                  bo - Bayesian optimisation, ea/ga: Evolutionary algorithm,
                  rand - Random search, direct: Dividing Rectangles, pdoo: PDOO
      config: Contains configuration parameters that are typically returned by
              exd.cp_domain_utils.load_config_file. config can be None only if domain
              is a EuclideanDomain object.
      options: Additional hyper-parameters for optimisation, as a namespace.
      reporter: A stream to print progress made during optimisation, or one of the
                following strings 'default', 'silent'. If 'silent', then it suppresses
                all outputs. If 'default', writes to stdout.
      * Alternatively, domain could be None if config is either a path_name to a
        configuration file or has configuration parameters.
    Returns:
      opt_val: The maximum value found during the optimisatio procdure.
      opt_pt: The corresponding optimum point.
      history: A record of the optimisation procedure which include the point evaluated
               and the values at each time step.
  """
  # Preprocess domain and config arguments
  raw_func = func
  domain, preproc_func_list, config, _ = preprocess_arguments(domain, [func], config)
  func = preproc_func_list[0]
  # Load arguments depending on domain type
  if domain.get_type() == 'euclidean':
    func_caller = EuclideanFunctionCaller(func, domain, vectorised=False)
  else:
    func_caller = CPFunctionCaller(func, domain, raw_func=raw_func,
                    domain_orderings=config.domain_orderings)
  # load options
  options = load_options_for_method(opt_method, 'opt', domain, capital_type, options)
  # Create worker manager and function caller
  worker_manager = get_worker_manager_from_capital_type(capital_type)
  # Optimise function here -----------------------------------------------------------
  if opt_method == 'bo':
    opt_val, opt_pt, history = gpb_from_func_caller(func_caller, worker_manager,
      max_capital, is_mf=False, options=options, reporter=reporter)
  elif opt_method in ['ga', 'ea']:
    opt_val, opt_pt, history = cp_ga_optimiser_from_proc_args(func_caller, domain,
      worker_manager, max_capital, options=options, reporter=reporter)
  elif opt_method == 'rand':
    opt_val, opt_pt, history = random_optimiser_from_func_caller(func_caller,
      worker_manager, max_capital, options=options, reporter=reporter)
  elif opt_method == 'direct':
    opt_val, opt_pt, history = direct_ft_maximise(func, domain.bounds, int(max_capital),
                                                  return_history=True)
  elif opt_method == 'pdoo':
    opt_val, opt_pt, history = pdoo_maximise_from_args(func, domain.bounds,
                                     int(max_capital), return_history=True)
  # Post processing
  if domain.get_type() == 'euclidean' and config is None:
    if opt_method not in ['direct', 'pdoo']:
      opt_pt = func_caller.get_raw_domain_coords(opt_pt)
      history.curr_opt_points = [func_caller.get_raw_domain_coords(pt)
                                 for pt in history.curr_opt_points]
      history.query_points = [func_caller.get_raw_domain_coords(pt)
                              for pt in history.query_points]
  else:
    opt_pt = get_raw_from_processed_via_config(opt_pt, config)
    history.curr_opt_points_raw = [get_raw_from_processed_via_config(pt, config)
                                   for pt in history.curr_opt_points]
    history.query_points_raw = [get_raw_from_processed_via_config(pt, config)
                                for pt in history.query_points]
  return opt_val, opt_pt, history


def minimise_function(func, *args, **kwargs):
  """
    Minimises a function func over domain domain. See maximise_function for a description
    of the arguments. All arguments are the same except for func, which should now be
    minimised.
  """
  func_to_max = lambda x: -func(x)
  max_val, opt_pt, history = maximise_function(func_to_max, *args, **kwargs)
  min_val = - max_val
  history = post_process_history_for_minimisation(history)
  return min_val, opt_pt, history


def minimise_multifidelity_function(func, *args, **kwargs):
  """
    Minimises a multifidelity function func over domain domain. See
    maximise_multifidelity_function for a description of the arguments. All arguments are
    the same except for func, which should now be minimised.
  """
  func_to_max = lambda x, z: -func(x, z)
  max_val, opt_pt, history = maximise_multifidelity_function(func_to_max, *args, **kwargs)
  min_val = - max_val
  history = post_process_history_for_minimisation(history)
  return min_val, opt_pt, history


# Alternative spelling
maximize_function = maximise_function
maximize_multifidelity_function = maximise_multifidelity_function
minimize_function = minimise_function
minimize_multifidelity_function = minimise_multifidelity_function




"""
  Adapted from pgmpy library: https://github.com/pgmpy/pgmpy
  Contains simple check functions
  -- kvysyara@andrew.cmu.edu
"""
import numpy as np


def _check_1d_array_object(parameter, name_param):
  """
  Checks whether given parameter is a 1d array like object,
  and returns a numpy array object
  """
  if isinstance(parameter, (np.ndarray, list, tuple, np.matrix)):
    parameter = np.array(parameter)
    if parameter.ndim != 1:
      raise TypeError("{} should be a 1d array type object".format(name_param))
  else:
    raise TypeError("{} should be a 1d array type object".format(name_param))

  return parameter


def _check_length_equal(param_1, param_2, name_param_1, name_param_2):
  """
  Raises an error when the length of given two arguments is not equal
  """
  if len(param_1) != len(param_2):
    raise ValueError("Length of {} must be same as Length of {}".\
                      format(name_param_1, name_param_2))



"""
  Unit tests for kernel.py
"""
from __future__ import absolute_import
from __future__ import division

# pylint: disable=import-error
# pylint: disable=no-member
# pylint: disable=invalid-name
# pylint: disable=relative-import
# pylint: disable=abstract-class-not-used
# pylint: disable=abstract-class-little-used

import numpy as np

# Local imports
from .kernel import Kernel, SEKernel, MaternKernel, PolyKernel, CoordinateProductKernel
from ..utils.ancillary_utils import dicts_are_equal
from ..utils.base_test_class import BaseTestClass, execute_tests


class KernelBasicTestCase(BaseTestClass):
  """ Basic tests for Kernel function """
  # pylint: disable=too-many-instance-attributes

  def setUp(self):
    """ Sets up attributes. """
    # Data for generic tests
    self.num_data_1 = 2
    self.num_data_2 = 3
    self.dim = 3
    self.hyper_param_dict_1 = {"param_1":1, "param_2":2, "param_3":3}
    self.hyper_param_dict_2 = {"param_4":4, "param_5":5, "param_6":6}
    self.hyper_param_dict_3 = {"param_1":1, "param_2":2, "param_3":3, "param_4":4,
                               "param_5":5, "param_6":6}
    # Data for the SE and poly kernels
    self.se_scale = 2
    self.data_1 = np.array([[1, 2], [3, 4.5]])
    self.data_2 = np.array([[1, 2], [3, 4]])
    # Data for the squared exponential kernel.
    self.dim_bandwidths = [0.1, 1]
    self.true_se_vals_11 = self.se_scale * np.array([[1, np.exp(-406.25/2)],
                                                     [np.exp(-406.25/2), 1]])
    self.true_se_vals_22 = self.se_scale * np.array([[1, np.exp(-404/2)],
                                                     [np.exp(-404/2), 1]])
    self.true_se_vals_12 = self.se_scale * np.array([[1, np.exp(-404/2)],
                                             [np.exp(-406.25/2), np.exp(-0.25/2)]])
    # Data for the Matern kernel
    self.nu_vals = [0.5, 1.5, 2.5]
    self.scaled_distances_11 = np.array([[0, np.sqrt(406.25)], [np.sqrt(406.25), 0]])
    self.scaled_distances_22 = np.array([[0, np.sqrt(404)], [np.sqrt(404), 0]])
    self.scaled_distances_12 = np.array([[0, np.sqrt(404)],
                                         [np.sqrt(406.25), np.sqrt(0.25)]])
    self.matern_scale = 2.1
    # Data for the polynomial kernel
    self.poly_order = 3
    self.poly_dim_scalings = [0.5, 2]
    self.poly_scale = 2
    self.true_poly_vals_11 = self.poly_scale * np.array([[17.25, 37.75],
                                                         [37.75, 84.25]])**self.poly_order
    self.true_poly_vals_22 = self.poly_scale * np.array([[17.25, 33.75],
                                                         [33.75, 67.25]])**self.poly_order
    self.true_poly_vals_12 = self.poly_scale * np.array([[17.25, 33.75],
                                                         [37.75, 75.25]])**self.poly_order
    # Data for the combined kernel
    self.com_scale = 4.3
    self.true_comb_11 = self.com_scale * self.true_se_vals_11 * self.true_poly_vals_11
    self.true_comb_22 = self.com_scale * self.true_se_vals_22 * self.true_poly_vals_22
    self.true_comb_12 = self.com_scale * self.true_se_vals_12 * self.true_poly_vals_12

  def test_basics(self):
    """ Tests basic functionality. """
    self.report('Testing basic functionality.')
    kern_1 = Kernel()
    kern_1.set_hyperparams(param_1=1, param_2=2, param_3=3)
    assert dicts_are_equal(kern_1.hyperparams, self.hyper_param_dict_1)
    kern_1.add_hyperparams(**self.hyper_param_dict_2)
    assert dicts_are_equal(kern_1.hyperparams, self.hyper_param_dict_3)
    kern_1.set_hyperparams(**self.hyper_param_dict_2)
    assert dicts_are_equal(kern_1.hyperparams, self.hyper_param_dict_2)

  def test_se_kernel(self):
    """ Tests for the SE kernel. """
    self.report('Tests for the SE kernel.')
    kern = SEKernel(self.data_1.shape[1], self.se_scale, self.dim_bandwidths)
    K11 = kern(self.data_1)
    K22 = kern(self.data_2)
    K12 = kern(self.data_1, self.data_2)
    assert np.linalg.norm(self.true_se_vals_11 - K11) < 1e-10
    assert np.linalg.norm(self.true_se_vals_22 - K22) < 1e-10
    assert np.linalg.norm(self.true_se_vals_12 - K12) < 1e-10

  @classmethod
  def _compute_matern_val_given_nu_scale(cls, nu, scale, dist):
    """ Computes the matern kernel given the num value. """
    if nu == 0.5:
      ret_K = np.exp(-dist)
    elif nu == 1.5:
      ret_K = np.exp(-np.sqrt(3) * dist) * (1 + np.sqrt(3) * dist)
    elif nu == 2.5:
      ret_K = np.exp(-np.sqrt(5) * dist) * (1 + np.sqrt(5) * dist + (5/3.0) * dist**2)
    else:
      raise ValueError('Implemented unit tests only for nu = 0.1, 1.5, 2.5')
    ret_K = scale * ret_K
    return ret_K

  def test_matern_kernel(self):
    """ Tests for the matern kernel. """
    self.report('Tests for the Matern kernel.')
    for nu in self.nu_vals:
      kern = MaternKernel(self.data_1.shape[1], nu, self.matern_scale,
                                 self.dim_bandwidths)
      K11 = (1.0/kern.norm_constant) * kern(self.data_1)
      K22 = (1.0/kern.norm_constant) * kern(self.data_2)
      K12 = (1.0/kern.norm_constant) * kern(self.data_1, self.data_2)
      true_vals_11 = self._compute_matern_val_given_nu_scale(nu, self.matern_scale,
                                                             self.scaled_distances_11)
      true_vals_22 = self._compute_matern_val_given_nu_scale(nu, self.matern_scale,
                                                             self.scaled_distances_22)
      true_vals_12 = self._compute_matern_val_given_nu_scale(nu, self.matern_scale,
                                                             self.scaled_distances_12)
      assert np.linalg.norm(true_vals_11 - K11) < 1e-10
      assert np.linalg.norm(true_vals_22 - K22) < 1e-10
      assert np.linalg.norm(true_vals_12 - K12) < 1e-10

  def test_poly_kernel(self):
    """ Tests for the polynomial kernel. """
    self.report('Tests for the Polynomial kernel.')
    kern = PolyKernel(self.data_1.shape[1], self.poly_order, self.poly_scale,
                             self.poly_dim_scalings)
    K11 = kern(self.data_1)
    K22 = kern(self.data_2)
    K12 = kern(self.data_1, self.data_2)
    assert np.linalg.norm(self.true_poly_vals_11 - K11) < 1e-10
    assert np.linalg.norm(self.true_poly_vals_22 - K22) < 1e-10
    assert np.linalg.norm(self.true_poly_vals_12 - K12) < 1e-10

  def test_comb_kernel(self):
    """ Tests for the combined kernel. """
    self.report('Tests for the Combined kernel.')
    kern_se = SEKernel(self.data_1.shape[1], self.se_scale, self.dim_bandwidths)
    kern_po = PolyKernel(self.data_1.shape[1], self.poly_order, self.poly_scale,
                                self.poly_dim_scalings)
    kern = CoordinateProductKernel(self.data_1.shape[0], self.com_scale,
                                          [kern_se, kern_po], [[0, 1], [0, 1]])
    K11 = kern(self.data_1)
    K22 = kern(self.data_2)
    K12 = kern(self.data_1, self.data_2)
    assert np.linalg.norm(self.true_comb_11 - K11) < 1e-10
    assert np.linalg.norm(self.true_comb_22 - K22) < 1e-10
    assert np.linalg.norm(self.true_comb_12 - K12) < 1e-10

  def test_effective_length_se(self):
    """ Tests for the effective length in the SE kernel. """
    # pylint: disable=too-many-locals
    self.report('Tests for the effective length in the SE kernel.')
    data_1 = np.array([1, 2])
    data_2 = np.array([[0, 1, 2], [1, 1, 0.5]])
    bws_1 = [0.1, 1]
    bws_2 = [0.5, 1, 2]
    res_l2_1 = np.sqrt(104)
    res_l2_2 = np.array([np.sqrt(2), np.sqrt(5.0625)])
    res_l1_1 = 12
    res_l1_2 = np.array([2, 3.25])
    dim_1 = 2
    dim_2 = 3
    all_data = [(data_1, bws_1, dim_1, res_l2_1, res_l1_1),
                (data_2, bws_2, dim_2, res_l2_2, res_l1_2)]
    for data in all_data:
      kern = SEKernel(data[2], 1, data[1])
      eff_l1_norms = kern.get_effective_norm(data[0], order=1,
                                             is_single=len(data[0].shape) == 1)
      eff_l2_norms = kern.get_effective_norm(data[0], order=2,
                                             is_single=len(data[0].shape) == 1)
      assert np.linalg.norm(eff_l2_norms - data[3]) < 1e-5
      assert np.linalg.norm(eff_l1_norms - data[4]) < 1e-5

  @classmethod
  def _compute_post_covar(cls, kern, X_tr, X_te):
    """ Computes the posterior covariance. """
    K_tr = kern.evaluate(X_tr, X_tr)
    K_tetr = kern.evaluate(X_te, X_tr)
    K_te = kern.evaluate(X_te, X_te)
    post_covar = K_te - K_tetr.dot(np.linalg.solve(K_tr, K_tetr.T))
    post_std = np.sqrt(np.diag(post_covar))
    return post_covar, post_std

  def test_compute_std_slack_se(self):
    """ Tests for the effective length in the SE kernel. """
    self.report('Tests for std slack in the SE kernel.')
    # The data here are in the order [dim, scale, num_data]
    prob_params = [[2, 1, 10], [3, 2, 0], [10, 6, 13]]
    n = 5
    for prob in prob_params:
      dim_bws = list(np.random.random(prob[0]) * 0.3 + 0.5)
      kern = SEKernel(prob[0], prob[1], dim_bws)
      X_1 = np.random.random((n, prob[0]))
      X_2 = np.random.random((n, prob[0]))
      X_tr = np.random.random((prob[2], prob[0]))
      _, std_1 = self._compute_post_covar(kern, X_tr, X_1)
      _, std_2 = self._compute_post_covar(kern, X_tr, X_2)
      std_diff = np.abs(std_1 - std_2)
      std_slack = kern.compute_std_slack(X_1, X_2)
      diff_12_scaled = kern.get_effective_norm(X_1 - X_2, order=2, is_single=False)
      kern_diff_12_scaled = kern.hyperparams['scale'] * diff_12_scaled
      assert np.all(std_diff <= std_slack)
      assert np.all(std_slack <= kern_diff_12_scaled)


if __name__ == '__main__':
  execute_tests()




"""
  A GA Optimiser for Cartesian Product Domains.
  We will use this mostly for optimising the acquisition in Dragonfly.
  -- kandasamy@cs.cmu.edu
"""

# pylint: disable=invalid-name

from copy import copy
import numpy as np
from scipy.spatial.distance import cdist
# Local imports
from ..exd.cp_domain_utils import get_processed_func_from_raw_func_for_cp_domain, \
                            load_cp_domain_from_config_file
from ..exd.experiment_caller import ExperimentCaller, CPFunctionCaller
from ..exd.exd_utils import get_cp_domain_initial_qinfos
from .ga_optimiser import GAOptimiser, ga_opt_args
from ..utils.general_utils import project_to_bounds
from ..utils.option_handler import load_options
from ..utils.reporters import get_reporter


cpga_opt_args = ga_opt_args


# Define Mutation Operators for each space individually
def _get_gauss_perturbation(x, bounds, sigmas=None):
  """ Returns the Gaussian perturbation. """
  if sigmas is None:
    sigmas = [(b[1] - b[0])/10 for b in bounds]
  epsilon = np.random.normal(scale=sigmas)
  return project_to_bounds(np.array(x) + epsilon, bounds)

def _return_ndarray_with_type(x, ret):
  """ Returns after checking type. """
  if isinstance(x, np.ndarray):
    return ret
  else:
    return type(x)(ret)

def euclidean_gauss_mutation(x, bounds, sigmas=None):
  """ Defines a Euclidean Mutation. """
  ret = _get_gauss_perturbation(x, bounds, sigmas)
  return _return_ndarray_with_type(x, ret)

def discrete_euclidean_mutation(x, list_of_items, uniform_prob=0.2):
  """ Makes a change depending on the vector values. """
  # cdist requires 2d input
  dists = cdist([x], list_of_items)[0]
  # Exponentiate and normalise to get the probabilities.
  unnorm_diff_probs = np.exp(-dists)
  sample_diff_probs = unnorm_diff_probs / unnorm_diff_probs.sum()
  # Change the distribution to interplate between it and uniform
  n = sample_diff_probs.shape[0]
  unif = np.full(sample_diff_probs.shape, 1. / n)
  p = (1 - uniform_prob) * sample_diff_probs + uniform_prob * unif
  # Now draw the samples
  idxs = np.arange(n)
  idx = np.random.choice(idxs, p=p)
  ret = list_of_items[idx]
  return _return_ndarray_with_type(x, ret)

def integral_gauss_mutation(x, bounds, sigmas=None):
  """ Defines a Euclidean Mutation. """
  ret = _get_gauss_perturbation(x, bounds, sigmas)
  ret = ret.round().astype(np.int)
  return _return_ndarray_with_type(x, ret)

def prod_discrete_random_mutation(x, list_of_list_of_items):
  """ Randomly chooses an index and changes it to something else from
      list_of_list_of_items. """
  ret = [copy(elem) for elem in x]
  change_idx = np.random.choice(len(x))
  change_list = copy(list_of_list_of_items[change_idx])
  change_list.remove(x[change_idx])
  change_val = np.random.choice(change_list)
  ret[change_idx] = change_val
  return ret

def discrete_random_mutation(x, list_of_items):
  """ Randomly chooses an index and changes it to something else from
      list_of_items. """
  ret = prod_discrete_random_mutation([x], [list_of_items])
  return ret[0]

def prod_discrete_numeric_exp_mutation(x, list_of_list_of_items, uniform_prob=0.2):
  """ Makes a change depending on the numerical values. """
  diff_vals = [np.abs(loi - x[idx]) for idx, loi in enumerate(list_of_list_of_items)]
  # Exponentiate and normalise to get the probabilities.
  unnorm_diff_probs = [np.exp(-elem) for elem in diff_vals]
  sample_diff_probs = [elem/elem.sum() for elem in unnorm_diff_probs]
  # Now change the distribution so that we pick one randomly with prob uniform_prob.
  sample_probs = [(1 - uniform_prob) * elem +
                  uniform_prob * np.ones((len(elem),)) / float(len(elem))
                  for elem in sample_diff_probs]
  # Now draw the samples
  ret = [np.random.choice(loi, p=p) for (loi, p) in
         zip(list_of_list_of_items, sample_probs)]
  return ret

def discrete_numeric_exp_mutation(x, list_of_items):
  """ Randomly chooses an index and changes it to something else from
      list_of_items. """
  ret = prod_discrete_numeric_exp_mutation([x], [list_of_items])
  return ret[0]


def get_default_mutation_op(dom):
  """ Returns the default mutation operator for the domain. """
  if dom.get_type() == 'euclidean':
    return lambda x: euclidean_gauss_mutation(x, dom.bounds)
  elif dom.get_type() == 'integral':
    return lambda x: integral_gauss_mutation(x, dom.bounds)
  elif dom.get_type() == 'discrete':
    return lambda x: discrete_random_mutation(x, dom.list_of_items)
  elif dom.get_type() == 'prod_discrete':
    return lambda x: prod_discrete_random_mutation(x, dom.list_of_list_of_items)
  elif dom.get_type() == 'discrete_numeric':
    return lambda x: discrete_numeric_exp_mutation(x, dom.list_of_items)
  elif dom.get_type() == 'prod_discrete_numeric':
    return lambda x: prod_discrete_numeric_exp_mutation(x, dom.list_of_list_of_items)
  elif dom.get_type() == 'discrete_euclidean':
    return lambda x: discrete_euclidean_mutation(x, dom.list_of_items)
  elif dom.get_type() == 'neural_network':
    from ..nn.nn_modifiers import get_single_nn_mutation_op
    return get_single_nn_mutation_op(dom, [0.5, 0.25, 0.125, 0.075, 0.05])
  else:
    raise ValueError('No default mutation implemented for domain type %s.'%(
                     dom.get_type()))


class CPGAOptimiser(GAOptimiser):
  """ A GA Optimiser for Cartesian Product Domains. """

  def __init__(self, func_caller, worker_manager, single_mutation_ops=None,
               single_crossover_ops=None, options=None, reporter=None):
    """ Constructor. """
    if options is None:
      reporter = get_reporter(reporter)
      options = load_options(cpga_opt_args, reporter=reporter)
    super(CPGAOptimiser, self).__init__(func_caller, worker_manager,
      mutation_op=self._mutation_op, crossover_op=self._crossover_op,
      options=options, reporter=reporter)
    self._set_up_single_mutation_ops(single_mutation_ops)
    self._set_up_single_crossover_ops(single_crossover_ops)

  def _set_up_single_mutation_ops(self, single_mutation_ops):
    """ Set up mutation operations. """
    if single_mutation_ops is None:
      single_mutation_ops = [None] * self.domain.num_domains
    for idx, dom in enumerate(self.domain.list_of_domains):
      if single_mutation_ops[idx] is None:
        single_mutation_ops[idx] = get_default_mutation_op(dom)
    self.single_mutation_ops = single_mutation_ops

  def _set_up_single_crossover_ops(self, crossover_ops):
    """ Set up cross-over operations. """
    # pylint: disable=unused-argument
    self.crossover_ops = crossover_ops

  def _mutation_op(self, X, num_mutations):
    """ The mutation operator for the product domain. """
    if hasattr(num_mutations, '__iter__'):
      num_mutations_for_each_x = num_mutations
    else:
      choices_for_each_mutation = np.random.choice(len(X), num_mutations, replace=True)
      num_mutations_for_each_x = [np.sum(choices_for_each_mutation == i) for i
                                  in range(len(X))]
    ret = []
    # Now extend
    for idx in range(len(X)):
      ret.extend(self._get_mutation_for_single_x(X[idx],
                 num_mutations_for_each_x[idx]))
    np.random.shuffle(ret)
    return ret

  def _get_mutation_for_single_x(self, x, num_mutations):
    """ Gets the mutation for single x. """
    ret = []
    for _ in range(num_mutations):
      curr_mutation = []
      for idx, elem in enumerate(x):
        curr_mutation.append(self.single_mutation_ops[idx](elem))
      ret.append(curr_mutation)
    return ret

  def _crossover_op(self):
    """ The crossover operation for the product domain. """
    raise NotImplementedError('Not implemented cross over operation yet.')

  # Define a function to obtain the initial qinfos =======================================
  def _get_initial_qinfos(self, num_init_evals):
    """ Gets num_init_evals initial points. """
    return get_cp_domain_initial_qinfos(self.domain, num_init_evals,
                                        dom_euclidean_sample_type='latin_hc',
                                        dom_integral_sample_type='latin_hc',
                                        dom_nn_sample_type='rand')


# APIs =================================================================================
def cp_ga_optimiser_from_proc_args(func_caller, cp_domain, worker_manager, max_capital,
                                   mode='asy', orderings=None, single_mutation_ops=None,
                                   single_crossover_ops=None, options=None,
                                   reporter=None):
  """ A GA optimiser on Cartesian product space from the function caller. """
  if not isinstance(func_caller, ExperimentCaller):
    func_caller = CPFunctionCaller(func_caller, cp_domain, domain_orderings=orderings)
  if options is None:
    reporter = get_reporter(reporter)
    options = load_options(cpga_opt_args, reporter=reporter)
  options.mode = mode
  from ..exd.worker_manager import RealWorkerManager, SyntheticWorkerManager
  if isinstance(worker_manager, RealWorkerManager):
    options.capital_type = 'realtime'
  elif isinstance(worker_manager, SyntheticWorkerManager):
    options.capital_type = 'return_value'
  return (CPGAOptimiser(func_caller, worker_manager,
                        single_mutation_ops=single_mutation_ops,
                        single_crossover_ops=single_crossover_ops,
                        options=options, reporter=reporter)).optimise(max_capital)


def cp_ga_optimiser_from_raw_args(raw_func, domain_config_file, worker_manager,
                                  max_capital, mode='asy', single_mutation_ops=None,
                                  single_crossover_ops=None, options=None,
                                  reporter='default'):
  """ Optimise a function from raw args such as the raw function and the configuration
      file. """
  # pylint: disable=no-member
  cp_dom, orderings = load_cp_domain_from_config_file(domain_config_file)
  proc_func = get_processed_func_from_raw_func_for_cp_domain(
                raw_func, cp_dom, orderings.index_ordering, orderings.dim_ordering)
  func_caller = CPFunctionCaller(proc_func, cp_dom, raw_func=raw_func,
                                 domain_orderings=orderings)
  return cp_ga_optimiser_from_proc_args(func_caller, cp_dom, worker_manager, max_capital,
                                        mode, orderings, single_mutation_ops,
                                        single_crossover_ops, options, reporter)




"""
  Implements some instances of a random optimiser for multi-objective optimisation.
  -- kandasamy@cs.cmu.edu
"""

# pylint: disable=invalid-name

from argparse import Namespace
import numpy as np
# Local
from ..exd.cp_domain_utils import get_processed_func_from_raw_func_for_cp_domain, \
                                load_cp_domain_from_config_file
from ..exd import domains
from ..exd.exd_utils import get_euclidean_initial_qinfos, get_cp_domain_initial_qinfos
from ..exd.experiment_caller import CPMultiFunctionCaller
from ..exd.cp_domain_utils import sample_from_cp_domain
from .multiobjective_optimiser import MultiObjectiveOptimiser, \
                                         multiobjective_opt_args
from ..utils.option_handler import load_options
from ..utils.reporters import get_reporter
from ..utils.general_utils import map_to_bounds


random_multiobjective_optimiser_args = multiobjective_opt_args
euclidean_random_multiobjective_optimiser_args = random_multiobjective_optimiser_args
cp_random_multiobjective_optimiser_args = random_multiobjective_optimiser_args


class RandomMultiObjectiveOptimiser(MultiObjectiveOptimiser):
  """ Optimises multiple objectives with random evaluations. """
  #pylint: disable=abstract-method

  def __init__(self, multi_func_caller, worker_manager, options=None, reporter=None):
    """ Constructor. """
    reporter = get_reporter(reporter)
    if options is None:
      options = load_options(random_multiobjective_optimiser_args, reporter=reporter)
    super(RandomMultiObjectiveOptimiser, self).__init__(multi_func_caller, worker_manager,
                                           model=None, options=options, reporter=reporter)

  def _multi_opt_method_set_up(self):
    """ Any set up specific to the method for multi-objective optimisation. """
    pass

  def _get_method_str(self):
    """ Returns a string describing the method. """
    return 'rand'

  def _add_data_to_model(self, qinfos):
    """ Adds data to model. """
    pass

  def _child_build_new_model(self):
    """ Builds a new model. """
    pass


class EuclideanRandomMultiObjectiveOptimiser(RandomMultiObjectiveOptimiser):
  """ A class which optimises for multiple objectives in Euclidean spaces using random
      evaluations.
  """

  def is_an_mf_method(self):
    """ Returns False since this is not a MF method. """
    return False

  def _determine_next_query(self):
    """ Determines the next query. """
    qinfo = Namespace(point=map_to_bounds(np.random.random(self.domain.dim),
                                          self.domain.bounds))
    return qinfo

  def _determine_next_batch_of_queries(self, batch_size):
    """ Determines the next batch of queries. """
    qinfos = [self._determine_next_query() for _ in range(batch_size)]
    return qinfos

  def _get_initial_qinfos(self, num_init_evals):
    """ Returns initial qinfos. """
    return get_euclidean_initial_qinfos(self.options.init_method, num_init_evals,
                                        self.domain.bounds)


class CPRandomMultiObjectiveOptimiser(RandomMultiObjectiveOptimiser):
  """ A class which optimises for multiple objectives in Cartesian product spaces using
      random evaluations.
  """

  def is_an_mf_method(self):
    """ Returns False since it is not a False method. """
    return False

  def _determine_next_query(self):
    """ Determines the next query. """
    qinfo = Namespace(point=sample_from_cp_domain(self.domain, 1)[0])
    return qinfo

  def _determine_next_batch_of_queries(self, batch_size):
    """ Determines the next batch of queries. """
    qinfos = [self._determine_next_query() for _ in range(batch_size)]
    return qinfos

  def _get_initial_qinfos(self, num_init_evals):
    """ Returns initial qinfos. """
    return get_cp_domain_initial_qinfos(self.domain, num_init_evals,
                                        dom_euclidean_sample_type='latin_hc',
                                        dom_integral_sample_type='latin_hc',
                                        dom_nn_sample_type='rand')


# APIs -----------------------------------------------------------------------------------
def random_multiobjective_optimisation_from_multi_func_caller(multi_func_caller,
    worker_manager, max_capital, mode, options=None, reporter='default'):
  """ Creates an appropriate MultiObjectiveOptimiser object and optimises the functions
      in multi_func_caller.
  """
  reporter = get_reporter(reporter)
  if isinstance(multi_func_caller.domain, domains.EuclideanDomain):
    moo_constructor = EuclideanRandomMultiObjectiveOptimiser
    dflt_list_of_options = euclidean_random_multiobjective_optimiser_args
  elif isinstance(multi_func_caller.domain, domains.CartesianProductDomain):
    moo_constructor = CPRandomMultiObjectiveOptimiser
    dflt_list_of_options = cp_random_multiobjective_optimiser_args
  else:
    raise ValueError('Random optimiser not implemented for domain of type %s.'%(
                     type(multi_func_caller.domain)))
  # Load options
  if options is None:
    options = load_options(dflt_list_of_options)
  options.mode = mode
  # Create Optimiser
  moo_optimiser = moo_constructor(multi_func_caller, worker_manager, options, reporter)
  # optimise and return
  return moo_optimiser.optimise(max_capital)


def cp_random_multiobjective_optimisation_from_raw_args(raw_funcs,
    domain_config_file, *args, **kwargs):
  """ A random multi-objective optimiser on CP spaces. """
  # pylint: disable=no-member
  cp_dom, orderings = load_cp_domain_from_config_file(domain_config_file)
  proc_funcs = [get_processed_func_from_raw_func_for_cp_domain(
                  rf, cp_dom, orderings.index_ordering, orderings.dim_ordering)
                for rf in raw_funcs]
  multi_func_caller = CPMultiFunctionCaller(proc_funcs, cp_dom, raw_funcs=raw_funcs,
                                            domain_orderings=orderings)
  return random_multiobjective_optimisation_from_multi_func_caller(multi_func_caller,
                                                                   *args, **kwargs)






"""
  Slice Sampler -- Adapted from pymc3 library: https://github.com/pymc-devs/pymc3
  -- kvysyara@andrew.cmu.edu
"""
from __future__ import division

# pylint: disable=invalid-name

import numpy as np
import numpy.random as nr

__all__ = ['Slice']

LOOP_ERR_MSG = 'max slicer iters %d exceeded'

class Slice(object):
  """
  Class for Slice sampler.

  Parameters
  ----------
  model : Distribution from which sampling has done
  w     : Initial width for slice
  tune  : Flag for tuning
  """

  def __init__(self, model, w=1., tune=True, iter_limit=np.inf):
    """
    Constructor for slice sampler.
    """
    self.model = model
    self.w = w
    self.tune = tune
    self.n_tunes = 0.
    self.iter_limit = iter_limit

  def _sample(self, q0):
    """
    Helper function which implements slice sampler.
    """
    self.w = np.resize(self.w, len(q0))  # this is a repmat
    q = np.copy(q0)
    ql = np.copy(q0)  # l for left boundary
    qr = np.copy(q0)  # r for right boudary
    for i, _ in enumerate(q0):
      # uniformly sample from 0 to p(q), but in log space
      y = self.model.logp(q) - nr.standard_exponential()
      ql[i] = q[i] - nr.uniform(0, self.w[i])
      qr[i] = q[i] + self.w[i]

      # Stepping out procedure
      cnt = 0
      while y < self.model.logp(ql):
        ql[i] -= self.w[i]
        cnt += 1
        if cnt > self.iter_limit:
          raise RuntimeError(LOOP_ERR_MSG % self.iter_limit)
      cnt = 0
      while y < self.model.logp(qr):
        qr[i] += self.w[i]
        cnt += 1
        if cnt > self.iter_limit:
          raise RuntimeError(LOOP_ERR_MSG % self.iter_limit)

      cnt = 0
      q[i] = (qr[i] - ql[i])*nr.rand() + ql[i]
      while self.model.logp(q) < y:
        # Sample uniformly from slice
        if q[i] > q0[i]:
          qr[i] = q[i]
        elif q[i] < q0[i]:
          ql[i] = q[i]
        q[i] = (qr[i] - ql[i])*nr.rand() + ql[i]
        cnt += 1
        if cnt > self.iter_limit:
          raise RuntimeError(LOOP_ERR_MSG % self.iter_limit)

      if self.tune:
        self.w[i] = self.w[i] * (self.n_tunes / (self.n_tunes + 1)) +\
          (qr[i] - ql[i]) / (self.n_tunes + 1)  # same as before
        qr[i] = q[i]
        ql[i] = q[i]

    if self.tune:
      self.n_tunes += 1

    return q

  def sample(self, q0, num_samples=1, burn=100):
    """
    sample -- function which populates samples
    by calling function _sample iteratively
    """
    if num_samples is None:
      num_samples = 1
    if not hasattr(q0, '__len__'):
      q0 = np.array([q0])

    # Burn in samples
    for i in range(burn):
      q0 = self._sample(q0)

    samples = np.zeros([num_samples, len(q0)])
    for i in range(num_samples):
      samples[i] = self._sample(q0)
      q0 = samples[i]

    return samples



"""
  Classes for various continuous probability distributions
  -- kvysyara@andrew.cmu.edu
"""
from __future__ import absolute_import
from __future__ import division

# pylint: disable=invalid-name
# pylint: disable=unused-argument
# pylint: disable=no-self-use

import numpy as np

# Local imports
from .distribution import Continuous
from ..exd import domains

# Univariate Normal Distribution
class Normal(Continuous):
  """ Univariate Normal Distribution """

  def __init__(self, mean, var):
    """ Constructor. """
    super(Normal, self).__init__()
    self.mean = float(mean)
    self.var = float(var)
    self.dim = 1
    self.domain = domains.EuclideanDomain([[-np.inf, np.inf]])

  def pdf(self, x):
    """ Returns value of pdf at x """
    return np.asscalar(np.exp(-((x-self.mean)**2)/(2*self.var))/\
                       (np.sqrt(2*np.pi*self.var)))

  def draw_samples_random(self, size=None):
    """ Returns random samples drawn from distribution. """
    return np.random.normal(self.mean, np.sqrt(self.var), size)

  def logp(self, x):
    """ Returns the log pdf at x """
    return np.asscalar(-0.5*(((x-self.mean)*(x-self.mean)/self.var)\
                        + np.log(2*np.pi*self.var)))

  def grad_logp(self, x):
    """ Returns gradient of log pdf at x """
    return np.asscalar(-(x - self.mean)/self.var)

  def get_mean(self):
    """ Returns mean. """
    return self.mean

  def get_variance(self):
    """ Returns variance. """
    return self.var

  def __str__(self):
    """ Returns a string representation. """
    return 'Univariate Normal: mean=%0.3f, variance=%.3f' % (self.mean, self.var)


# Multivariate Gaussian Distribution
class MultivariateGaussian(Continuous):
  """ Multivariate Gaussian Distribution """

  def __init__(self, mean, cov):
    """ Constructor. """
    super(MultivariateGaussian, self).__init__()
    self.mean = np.array(mean, dtype=float)
    self.cov = np.array(cov, dtype=float)
    self.pre = np.linalg.inv(self.cov)
    self.det = np.linalg.det(self.cov)
    self.dim = len(mean)
    self.domain = domains.EuclideanDomain(np.tile(np.array([-np.inf, np.inf]),\
                                          (len(mean), 1)))

  def pdf(self, x):
    """ Returns value of pdf at x """
    value = -0.5*np.dot(np.transpose(x - self.mean), np.dot(self.pre, x - self.mean))
    return np.asscalar(np.exp(value)*np.power(2*np.pi*self.det, -0.5))

  def draw_samples_random(self, size):
    """ Returns random samples drawn from distribution. """
    return np.random.multivariate_normal(self.mean, self.cov, size)

  def logp(self, x):
    """ Returns the log pdf at x """
    value = np.dot(np.transpose(x - self.mean), np.dot(self.pre, x - self.mean))
    return -0.5*(value + np.log(2*np.pi) + np.log(self.det))

  def grad_logp(self, x):
    """ Returns gradient of log pdf at x """
    return -np.dot(self.pre, x - self.mean)

  def get_mean(self):
    """ Returns mean. """
    return self.mean

  def get_variance(self):
    """ Returns variance. """
    return self.cov

  def __str__(self):
    """ Returns a string representation. """
    return 'Multivariate Normal'


# Continuous Uniform Distribution
class ContinuousUniform(Continuous):
  """ Continuous Uniform Distribution """

  def __init__(self, lower, upper):
    """ Constructor. """
    super(ContinuousUniform, self).__init__()
    self.lower = float(lower)
    self.upper = float(upper)
    self.dim = 1
    self.domain = domains.EuclideanDomain([[self.lower, self.upper]])

  def pdf(self, x):
    """ Returns value of pdf at x """
    if x < self.lower or x > self.upper:
      return 0
    return np.asscalar(1/(self.upper - self.lower))

  def draw_samples_random(self, size=None):
    """ Returns random samples drawn from distribution. """
    return np.random.uniform(self.lower, self.upper, size)

  def logp(self, x):
    """ Returns the log pdf at x """
    if x < self.lower or x > self.upper:
      return -np.inf
    return -np.log(self.upper - self.lower)

  def grad_logp(self, x):
    """ Returns gradient of log pdf at x """
    return 0

  def get_parameters(self):
    """ Returns parameters """
    return self.lower, self.upper

  def get_mean(self):
    """ Returns mean """
    return (self.lower + self.upper)/2

  def get_variance(self):
    """ Returns variance """
    return (np.power((self.upper - self.lower), 2))/12

  def __str__(self):
    """ Returns a string representation. """
    return 'Continuous Uniform: lower=%0.3f, upper=%.3f' % (self.lower, self.upper)


# Exponential Distribution
class Exponential(Continuous):
  """ Exponential Distribution """

  def __init__(self, lam):
    """ Constructor. """
    super(Exponential, self).__init__()
    self.lam = float(lam)
    self.dim = 1
    self.domain = domains.EuclideanDomain([[0, np.inf]])

  def pdf(self, x):
    """ Returns value of pdf at x """
    if x < 0:
      return 0
    return np.asscalar(self.lam*np.exp(-self.lam*x))

  def draw_samples_random(self, size=None):
    """ Returns random samples drawn from distribution. """
    return np.random.exponential(1/self.lam, size)

  def logp(self, x):
    """ Returns the log pdf at x """
    if x < 0:
      return -np.inf
    return np.log(self.lam) - self.lam*x

  def grad_logp(self, x):
    """ Returns gradient of log pdf at x """
    return -self.lam

  def get_lambda(self):
    """ Returns lambda parameter"""
    return self.lam

  def get_mean(self):
    """ Returns mean """
    return 1/self.lam

  def get_variance(self):
    """ Returns variance """
    return np.power(self.lam, -2)

  def __str__(self):
    """ Returns a string representation. """
    return 'Exponential: lambda=%0.3f' % (self.lam)


# Beta Distribution
class Beta(Continuous):
  """ Beta Distribution """

  def __init__(self, alpha, beta):
    """ Constructor. """
    super(Beta, self).__init__()
    self.alpha = float(alpha)
    self.beta = float(beta)
    self.dim = 1
    self.B = (np.math.factorial(self.alpha - 1)*np.math.factorial(self.beta - 1))/\
                  (float(np.math.factorial(self.alpha+self.beta-1)))
    self.domain = domains.EuclideanDomain([[0, 1]])

  def pdf(self, x):
    """ Returns value of pdf at x """
    if x < 0 or x > 1:
      return 0
    return np.asscalar((np.power(x, self.alpha - 1)*\
                        np.power(1 - x, self.beta - 1))/self.B)

  def draw_samples_random(self, *args):
    """ Returns random samples drawn from distribution. """
    return np.random.beta(self.alpha, self.beta, *args)

  def logp(self, x):
    """ Returns the log pdf at x """
    if x < 0 or x > 1:
      return -np.inf
    return (self.alpha - 1)*np.log(x) + (self.beta - 1)*np.log(1-x) - np.log(self.B)

  def grad_logp(self, x):
    """ Returns gradient of log pdf at x """
    if x < 0 or x > 1:
      return 0
    return (self.alpha - 1)/x - (self.beta - 1)/(1 - x)

  def get_parameters(self):
    """ Returns parameters """
    return self.alpha, self.beta

  def get_mean(self):
    """ Returns mean """
    return self.alpha/(self.alpha + self.beta)

  def get_variance(self):
    """ Returns variance """
    return (self.alpha*self.beta)/\
            ((np.power(self.alpha+self.beta, 2))*(self.alpha+self.beta+1))

  def __str__(self):
    """ Returns a string representation. """
    return 'Beta: alpha=%0.3f, beta=%.3f' % (self.alpha, self.beta)



"""
  Test cases for functions in ancillary_utils.py
  -- kandasamy@cs.cmu.edu
"""
# pylint: disable=import-error
# pylint: disable=no-member
# pylint: disable=invalid-name
# pylint: disable=relative-import

from . import ancillary_utils
from .base_test_class import BaseTestClass, execute_tests


class AncillaryUtilsTestCase(BaseTestClass):
  """ Unit test class for ancillary utilities. """

  def setUp(self):
    """ Sets up attributes. """
    # For dictionary
    self.dict_1 = {'a':1, 'b':2, 'c':3}
    self.dict_2 = {'b':2, 'a':1, 'c':3}
    self.dict_3 = {'a':4, 'b':5, 'c':6}
    self.dict_4 = {'d':1, 'e':2, 'f':3}
    # Lists
    self.list_1 = [1, 2, 3, 9, 10]
    self.list_2 = [1, 3, 3, 4, 19]
    self.list_3 = [1, 5, 8, 3]
    self.list_4 = [-x for x in self.list_1]
    self.list_5 = [-x for x in self.list_2]

  def test_compare_dict(self):
    """ Test compare_dict function. """
    self.report('dicts_are_equal')
    assert ancillary_utils.dicts_are_equal(self.dict_1, self.dict_2)
    assert not ancillary_utils.dicts_are_equal(self.dict_1, self.dict_3)
    assert not ancillary_utils.dicts_are_equal(self.dict_1, self.dict_4)

  def test_nondecreasing(self):
    """ Unit tests for is_nondecreasing. """
    assert ancillary_utils.is_nondecreasing(self.list_1)
    assert ancillary_utils.is_nondecreasing(self.list_2)
    assert not ancillary_utils.is_nondecreasing(self.list_3)
    assert not ancillary_utils.is_nondecreasing(self.list_4)
    assert not ancillary_utils.is_nondecreasing(self.list_5)

  def test_nonincreasing(self):
    """ Unit tests for is_nonincreasing. """
    assert not ancillary_utils.is_nonincreasing(self.list_1)
    assert not ancillary_utils.is_nonincreasing(self.list_2)
    assert not ancillary_utils.is_nonincreasing(self.list_3)
    assert ancillary_utils.is_nonincreasing(self.list_4)
    assert ancillary_utils.is_nonincreasing(self.list_5)

  def test_decreasing(self):
    """ Unit tests for is_decreasing. """
    assert not ancillary_utils.is_decreasing(self.list_1)
    assert not ancillary_utils.is_decreasing(self.list_2)
    assert not ancillary_utils.is_decreasing(self.list_3)
    assert ancillary_utils.is_decreasing(self.list_4)
    assert not ancillary_utils.is_decreasing(self.list_5)

  def test_increasing(self):
    """ Unit tests for is_increasing. """
    assert ancillary_utils.is_increasing(self.list_1)
    assert not ancillary_utils.is_increasing(self.list_2)
    assert not ancillary_utils.is_increasing(self.list_3)
    assert not ancillary_utils.is_increasing(self.list_4)
    assert not ancillary_utils.is_increasing(self.list_5)


if __name__ == '__main__':
  execute_tests()




"""
  Multi Objective Acquisition functions for Bayesian Optimisation.
  - bparia@cs.cmu.edu
"""

from __future__ import division

# pylint: disable=invalid-name

from argparse import Namespace
from copy import copy
import numpy as np
# Local
from .gpb_acquisitions import maximise_acquisition
from .gpb_acquisitions import get_gp_sampler_for_parallel_strategy


# Multi-Objective Thompson Sampling ========================================
def mo_lin_asy_ts(gps, anc_data):
  """ Returns a recommendation via TS with Linear Scalarization in the
  asynchronous setting for MOO """
  anc_data = copy(anc_data)
  # Always use a random optimiser with a vectorised sampler for TS.
  if anc_data.acq_opt_method != 'rand':
    anc_data.acq_opt_method = 'rand'
    anc_data.max_evals = 4 * anc_data.max_evals
  gp_samples = []
  for gp in gps:
    gp_sample = get_gp_sampler_for_parallel_strategy(gp, anc_data)
    gp_samples.append(gp_sample)

  def acquisition(x):
    """ Computes the scalar acquisition function for the given weight
        vector and the reference point.
    """
    s = 0.0
    for gp_sample, weight in zip(gp_samples, anc_data.obj_weights):
      s += gp_sample(x) * weight
    return s

  return maximise_acquisition(acquisition, anc_data, vectorised=True)


def mo_tch_asy_ts(gps, anc_data):
  """ Returns a recommendation via TS with Tchebychev Scalarization in the
  asynchronous setting for MOO """
  anc_data = copy(anc_data)
  # Always use a random optimiser with a vectorised sampler for TS.
  if anc_data.acq_opt_method != 'rand':
    anc_data.acq_opt_method = 'rand'
    anc_data.max_evals = 4 * anc_data.max_evals
  gp_samples = []
  for gp in gps:
    gp_sample = get_gp_sampler_for_parallel_strategy(gp, anc_data)
    gp_samples.append(gp_sample)
  # TS acquisition
  def acquisition(x):
    """ Computes the scalar acquisition function for the given weight
        vector and the reference point.
    """
    s = np.full((len(x), ), np.inf)
    for gp_sample, weight, ref in \
        zip(gp_samples, anc_data.obj_weights, anc_data.reference_point):
      s = np.minimum(s, (gp_sample(x) - ref) / weight)
    return s
  # return
  return maximise_acquisition(acquisition, anc_data, vectorised=True)


# Multi-Objective UCB ========================================
def _get_ucb_beta_th(dim, time_step):
  """ Computes the beta t for UCB based methods. """
  return np.sqrt(0.2 * dim * np.log(2 * dim * time_step + 1))


def mo_lin_asy_ucb(gps, anc_data):
  """ Returns a recommendation via UCB with Linear Scalarization in the
  asynchronous setting for MOO """
  beta_th = _get_ucb_beta_th(anc_data.domain.dim, anc_data.t)
  def acquisition(x):
    """ Computes the GP-UCB acquisition. """
    mu_tot = 0.0
    sigma2_tot = 0.0
    for gp, weight in zip(gps, anc_data.obj_weights):
      mu, sigma = gp.eval(x, uncert_form='std')
      mu_tot += mu * weight
      sigma2_tot += sigma * sigma * weight**2
    ret = mu_tot + beta_th * np.sqrt(sigma2_tot)
    return ret
  return maximise_acquisition(acquisition, anc_data)


def mo_tch_asy_ucb(gps, anc_data):
  """ Returns a recommendation via UCB with Linear Scalarization in the
  asynchronous setting for MOO """
  beta_th = _get_ucb_beta_th(anc_data.domain.dim, anc_data.t)
  def acquisition(x):
    """ Computes the GP-UCB acquisition. """
    ret = np.asarray([np.inf for _ in range(len(x))])
    for gp, weight, ref in \
        zip(gps, anc_data.obj_weights, anc_data.reference_point):
      mu, sigma2 = gp.eval(x, uncert_form='std')
      ucb = mu + beta_th * np.sqrt(sigma2) - ref
      ret = np.minimum(ret, ucb / weight)
    return ret
  return maximise_acquisition(acquisition, anc_data)


asy = Namespace(
  lin_ts=mo_lin_asy_ts,
  tch_ts=mo_tch_asy_ts,
  lin_ucb=mo_lin_asy_ucb,
  tch_ucb=mo_tch_asy_ucb,
)

# TODO: implement synchronous versions. Need to implement a version of
# _get_syn_recommendations_from_asy in gpb_acquisitions.
syn = Namespace()

seq = Namespace(
  lin_ts=mo_lin_asy_ts,
  tch_ts=mo_tch_asy_ts,
  lin_ucb=mo_lin_asy_ucb,
  tch_ucb=mo_tch_asy_ucb,
)

